import os
import subprocess
from os.path import expanduser
from urllib import request

import boto3
import botocore
import click
import yaml
from kubernetes import client as kube_client
from kubernetes import config as kube_config
from kubernetes.client.rest import ApiException
from kubernetes.config.kube_config import KUBE_CONFIG_DEFAULT_LOCATION

CF_TEMPLATE_URL = "https://amazon-eks.s3-us-west-2.amazonaws.com/cloudformation/2020-08-12/amazon-eks-vpc-sample.yaml"
AWS_NVIDIA_DAEMONSET = "https://raw.githubusercontent.com/NVIDIA/k8s-device-plugin/v0.6.0/nvidia-device-plugin.yml"


def _create_ssh_key(ssh_key):

    # create ssh key

    ec2 = boto3.client("ec2")
    try:
        keypair = ec2.create_key_pair(KeyName=ssh_key)
        file_location = expanduser("~") + "/.ssh/{}.pem".format(ssh_key)
        click.echo("Writing ssh key to " + file_location)
        with open(file_location, "w") as file:
            file.write(keypair["KeyMaterial"])
    except botocore.exceptions.ClientError as error:
        if error.response["Error"]["Code"] == "InvalidKeyPair.Duplicate":

            # the key has already been created

            pass
        else:
            click.UsageError(error)


def _create_eks_nodegroup(
    name,
    num_workers,
    machine_type,
    nodeGroupName,
    ssh_key,
    ami_id,
    secGroupId,
    vpcId,
    subnets,
):

    params = [
        {"ParameterKey": "KeyName", "ParameterValue": ssh_key},
        {"ParameterKey": "NodeImageId", "ParameterValue": ami_id},
        {"ParameterKey": "NodeInstanceType", "ParameterValue": machine_type},
        {
            "ParameterKey": "NodeAutoScalingGroupMinSize",
            "ParameterValue": str(num_workers),
        },
        {
            "ParameterKey": "NodeAutoScalingGroupMaxSize",
            "ParameterValue": str(num_workers),
        },
        {
            "ParameterKey": "NodeAutoScalingGroupDesiredCapacity",
            "ParameterValue": str(num_workers),
        },
        {"ParameterKey": "ClusterName", "ParameterValue": name},
        {"ParameterKey": "NodeGroupName", "ParameterValue": nodeGroupName},
        {
            "ParameterKey": "ClusterControlPlaneSecurityGroup",
            "ParameterValue": secGroupId,
        },
        {"ParameterKey": "VpcId", "ParameterValue": vpcId},
        {"ParameterKey": "Subnets", "ParameterValue": ",".join(subnets)},
    ]

    stackName = "eks-auto-scaling-group-" + name
    templateURL = "https://amazon-eks.s3.us-west-2.amazonaws.com/cloudformation/2020-08-12/amazon-eks-nodegroup.yaml"

    cloudFormation = boto3.client("cloudformation")
    cloudFormation.create_stack(
        StackName=stackName,
        TemplateURL=templateURL,
        Parameters=params,
        Capabilities=["CAPABILITY_IAM"],
    )

    waiter = cloudFormation.get_waiter("stack_create_complete")
    click.echo("Waiting for nodegroup to be created.")
    waiter.wait(StackName=stackName)

    resource = boto3.resource("cloudformation")
    stack = resource.Stack(stackName)

    nodeInstanceRole = None
    for i in stack.outputs:
        if i["OutputKey"] == "NodeInstanceRole":
            nodeInstanceRole = i["OutputValue"]

    if nodeInstanceRole is None:
        raise ValueError("Could not get NodeInstanceRole")
    return (nodeInstanceRole, stackName)


def _create_config_map(nodeInstanceRole):

    # create config map to connect the nodes to the cluster

    v1 = kube_client.CoreV1Api()
    namespace = "kube-system"
    configMapName = "aws-auth"

    # Delete old config maps in case it exists

    body = kube_client.V1DeleteOptions()
    body.api_version = "v1"
    try:
        v1.delete_namespaced_config_map(
            name="aws-auth", namespace="kube-system", body=body
        )
    except ApiException as e:
        pass

    # Create new config map

    body = kube_client.V1ConfigMap()
    body.api_version = "v1"
    body.metadata = {}
    body.metadata["name"] = configMapName
    body.metadata["namespace"] = namespace

    body.data = {}
    body.data["mapRoles"] = (
        "- rolearn: "
        + nodeInstanceRole
        + """
  username: system:node:{{EC2PrivateDNSName}}
  groups:
    - system:bootstrappers
    - system:nodes
"""
    )

    response = v1.create_namespaced_config_map(namespace, body)


def aws_create_cluster(
    name,
    nodeGroupName,
    num_workers,
    machine_type,
    ssh_key,
    ami_id,
    kubernetes_version,
):

    assert num_workers >= 2, "Number of workers should be at least 2."

    # create VPC stack for cluster

    stack_name = name + "-stack"
    cf_client = boto3.client("cloudformation")
    cf_client.create_stack(StackName=stack_name, TemplateURL=CF_TEMPLATE_URL)
    waiter = cf_client.get_waiter("stack_create_complete")
    click.echo("Waiting for the VPC stack to be created.")

    # will throw an exception if the creation fails

    waiter.wait(StackName=stack_name)

    # obtain vpc id, security group id and subnet ids

    r = cf_client.describe_stack_resources(
        StackName=stack_name, LogicalResourceId="VPC"
    )
    vpcId = r["StackResources"][0]["PhysicalResourceId"]

    r = cf_client.describe_stack_resources(
        StackName=stack_name, LogicalResourceId="ControlPlaneSecurityGroup"
    )
    secGroupId = r["StackResources"][0]["PhysicalResourceId"]

    ec2 = boto3.resource("ec2")
    vpc = ec2.Vpc(vpcId)
    subnets = [subnet.id for subnet in vpc.subnets.all()]

    ec2 = boto3.client("ec2")

    # ensure that public IP addresses are assigned on launch in each subnet

    for subnet in subnets:
        ec2.modify_subnet_attribute(
            MapPublicIpOnLaunch={"Value": True}, SubnetId=subnet
        )

    # create a role and assign the policy needed for creating the EKS cluster

    iam = boto3.client("iam")
    try:
        assume_role_policy_document = """{
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"Service": "eks.amazonaws.com"},
                    "Action": "sts:AssumeRole"
                }
            ]
        }"""
        iam.create_role(
            RoleName="EKSClusterRole",
            AssumeRolePolicyDocument=assume_role_policy_document,
        )
    except botocore.exceptions.ClientError as error:
        if error.response["Error"]["Code"] == "EntityAlreadyExists":

            # the role has already been created

            pass
        else:
            click.UsageError(error)

    iam.attach_role_policy(
        RoleName="EKSClusterRole",
        PolicyArn="arn:aws:iam::aws:policy/AmazonEKSClusterPolicy",
    )

    roleArn = iam.get_role(RoleName="EKSClusterRole")["Role"]["Arn"]

    # create the EKS cluster

    eks = boto3.client("eks")
    eks.create_cluster(
        name=name,
        version=kubernetes_version,
        roleArn=roleArn,
        resourcesVpcConfig={"subnetIds": subnets, "securityGroupIds": [secGroupId]},
    )
    waiter = eks.get_waiter("cluster_active")
    click.echo("Waiting for cluster to be created. This can take up to ten minutes.")
    waiter.wait(name=name)

    (kube_context, cluster) = _create_kube_config_aws_entry(name)
    kube_config.load_kube_config(context=kube_context)

    # Create ssh key

    _create_ssh_key(ssh_key)

    # Create eks nodegroup

    (nodeInstanceRole, stackName) = _create_eks_nodegroup(
        name,
        num_workers,
        machine_type,
        nodeGroupName,
        ssh_key,
        ami_id,
        secGroupId,
        vpcId,
        subnets,
    )

    _create_config_map(nodeInstanceRole)

    return (kube_context, cf_client, stackName, cluster)


def _create_kube_config_aws_entry(name):
    eks = boto3.client("eks")
    cluster = eks.describe_cluster(name=name)
    cluster_cert = cluster["cluster"]["certificateAuthority"]["data"]
    cluster_ep = cluster["cluster"]["endpoint"]
    cluster_arn = cluster["cluster"]["arn"]

    cluster_config = {
        "apiVersion": "v1",
        "kind": "Config",
        "clusters": [
            {
                "cluster": {
                    "server": str(cluster_ep),
                    "certificate-authority-data": str(cluster_cert),
                },
                "name": cluster_arn,
            }
        ],
        "contexts": [
            {
                "context": {"cluster": cluster_arn, "user": cluster_arn},
                "name": cluster_arn,
            }
        ],
        "current-context": cluster_arn,
        "preferences": {},
        "users": [
            {
                "name": cluster_arn,
                "user": {
                    "exec": {
                        "apiVersion": "client.authentication.k8s.io/v1alpha1",
                        "command": "aws-iam-authenticator",
                        "args": ["token", "-i", name],
                    }
                },
            }
        ],
    }

    config_text = yaml.dump(cluster_config, default_flow_style=False)
    default_config = os.path.abspath(os.path.expanduser(KUBE_CONFIG_DEFAULT_LOCATION))
    config_dir = os.path.dirname(default_config)
    config_file = os.path.join(config_dir, "mlbench_aws")

    with open(config_file, "w") as f:
        f.write(config_text)

    shell_env = os.environ.copy()
    shell_env["KUBECONFIG"] = "{}:{}".format(default_config, config_file)

    with open(default_config, "w") as out:
        p = subprocess.Popen(
            ["kubectl", "config", "view", "--flatten"],
            stdout=out,
            stderr=subprocess.PIPE,
            env=shell_env,
        )
        p.communicate()

    return (cluster_arn, cluster)


def deploy_nvidia_daemonset_aws():
    """ Deploys the NVIDIA daemon set to the AWS cluster"""
    with request.urlopen(AWS_NVIDIA_DAEMONSET) as r:
        dep = yaml.safe_load(r)
        dep["spec"]["selector"] = {
            "matchLabels": dep["spec"]["template"]["metadata"]["labels"]
        }
        dep = kube_client.ApiClient()._ApiClient__deserialize(dep, "V1DaemonSet")
        k8s_client = kube_client.AppsV1Api()
        k8s_client.create_namespaced_daemon_set("kube-system", body=dep)


def setup_aws_kube_client(name):

    # Keeping this function so we can later configure the python kube client directly

    (kube_context, _) = _create_kube_config_aws_entry(name)
    return kube_context



# -*- coding: utf-8 -*-

"""Console script for mlbench_cli."""
import configparser
import json
import os
import pickle
import subprocess
import sys
from os.path import expanduser
from pathlib import Path

import boto3
import botocore.exceptions
import click
import matplotlib.pyplot as plt
import urllib3
import yaml
from appdirs import user_data_dir
from kubernetes import config as kube_config
from tabulate import tabulate

import mlbench_core
from mlbench_core.api import MLBENCH_BACKENDS, MLBENCH_IMAGES, ApiClient
from mlbench_core.cli.aws_utils import (
    aws_create_cluster,
    deploy_nvidia_daemonset_aws,
    setup_aws_kube_client,
)
from mlbench_core.cli.gcloud_utils import (
    deploy_nvidia_daemonset,
    gcloud_create_cluster,
    gcloud_delete_cluster,
    setup_gcloud_kube_client,
)
from mlbench_core.cli.kind_utils import (
    create_local_registry,
    kind_create_cluster,
    kind_delete_cluster,
)
from mlbench_core.cli.utils import deploy_chart

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


@click.group()
@click.version_option(mlbench_core.__version__, help="Print mlbench version")
def cli_group(args=None):
    """Console script for mlbench_cli."""
    return 0


@cli_group.command()
@click.argument("name", type=str)
@click.argument("num_workers", nargs=-1, type=int, metavar="num-workers")
@click.option("--gpu", "-g", default=False, type=bool, is_flag=True)
@click.option("--num-cpus", "-c", default=4, type=int)
@click.option("--light", "-l", default=False, type=bool, is_flag=True)
@click.option("--dashboard-url", "-u", default=None, type=str)
def run(name, num_workers, gpu, num_cpus, light, dashboard_url):
    """Start a new run for a benchmark image"""
    current_run_inputs = {}

    last_run_inputs_dir_location = os.path.join(
        os.environ["HOME"], ".local", "share", "mlbench"
    )
    Path(last_run_inputs_dir_location).mkdir(parents=True, exist_ok=True)

    last_run_inputs_file_location = os.path.join(
        last_run_inputs_dir_location, "last_run_inputs.pkl"
    )

    try:
        last_run_inputs = pickle.load(open(last_run_inputs_file_location, "rb"))
    except FileNotFoundError as e:
        last_run_inputs = {}

    images = list(MLBENCH_IMAGES.keys())

    text_prompt = "Benchmark: \n\n"

    text_prompt += "\n".join("[{}]\t{}".format(i, t) for i, t in enumerate(images))
    text_prompt += "\n[{}]\tCustom Image".format(len(images))

    text_prompt += "\n\nSelection"

    selection = click.prompt(
        text_prompt,
        type=click.IntRange(0, len(images)),
        default=last_run_inputs.get("benchmark", 0),
    )
    current_run_inputs["benchmark"] = selection

    if selection == len(images):
        # run custom image
        image = click.prompt(
            "Image", type=str, default=last_run_inputs.get("image", None)
        )
        current_run_inputs["image"] = image
        image_command = click.prompt(
            "Command", type=str, default=last_run_inputs.get("image_command", None)
        )
        current_run_inputs["image_command"] = image_command
        benchmark = {
            "custom_image_name": image,
            "custom_image_command": image_command,
        }
    else:
        benchmark = {"image": images[selection]}

    # Backend Prompt
    text_prompt = "Backend: \n\n"
    text_prompt += "\n".join(
        "[{}]\t{}".format(i, t) for i, t in enumerate(MLBENCH_BACKENDS)
    )
    text_prompt += "\n[{}]\tCustom Backend".format(len(MLBENCH_BACKENDS))
    text_prompt += "\n\nSelection"

    selection = click.prompt(
        text_prompt,
        type=click.IntRange(0, len(MLBENCH_BACKENDS)),
        default=last_run_inputs.get("backend", 0),
    )
    current_run_inputs["backend"] = selection

    if selection == len(MLBENCH_BACKENDS):
        backend = click.prompt(
            "Backend", type=str, default=last_run_inputs.get("custom_backend", None)
        )
        current_run_inputs["custom_backend"] = backend
        run_on_all = click.confirm(
            "Run command on all nodes (otherwise just first node)",
            default=last_run_inputs.get("run_on_all", None),
        )
        current_run_inputs["run_on_all"] = run_on_all
        benchmark["custom_backend"] = backend
        benchmark["run_all_nodes"] = run_on_all
    else:
        benchmark["backend"] = MLBENCH_BACKENDS[selection]

    pickle.dump(current_run_inputs, open(last_run_inputs_file_location, "wb"))

    benchmark["gpu_enabled"] = gpu
    benchmark["light_target"] = light
    benchmark["num_cpus"] = num_cpus - 1

    loaded = setup_client_from_config()

    client = ApiClient(in_cluster=False, url=dashboard_url, load_config=not loaded)

    results = []

    for num_w in num_workers:
        current_name = "{}-{}".format(name, num_w)

        res = client.create_run(current_name, num_w, **benchmark)
        results.append(res)

    for res in results:
        act_result = res.result()
        if act_result.status_code > 201:
            try:
                click.echo(
                    "Couldn't start run: {}".format(act_result.json()["message"])
                )
            except json.JSONDecodeError:
                print(str(act_result.text))
                click.echo(
                    "Couldn't start run: Status {} for request".format(
                        act_result.status_code
                    )
                )
            return

        click.echo("Run started with name {}".format(act_result.json()["name"]))


@cli_group.command()
@click.argument("name", type=str, required=False)
@click.option("--dashboard-url", "-u", default=None, type=str)
def status(name, dashboard_url):
    """Get the status of a benchmark run, or all runs if no name is given"""
    loaded = setup_client_from_config()

    client = ApiClient(in_cluster=False, url=dashboard_url, load_config=not loaded)

    ret = client.get_runs()
    runs = ret.result().json()

    if name is None:  # List all runs
        for run in runs:
            del run["job_id"]
            del run["job_metadata"]

        click.echo(tabulate(runs, headers="keys"))
        return

    try:
        run = next(r for r in runs if r["name"] == name)
    except StopIteration:
        click.echo("Run not found")
        return

    del run["job_id"]
    del run["job_metadata"]

    click.echo(tabulate([run], headers="keys"))

    loss = client.get_run_metrics(
        run["id"], metric_filter="val_global_loss @ 0", last_n=1
    )
    prec = client.get_run_metrics(
        run["id"], metric_filter="val_global_Prec@1 @ 0", last_n=1
    )

    loss = loss.result()
    prec = prec.result()

    if loss.status_code < 300 and "val_global_loss @ 0" in loss.json():
        val = loss.json()["val_global_loss @ 0"][0]
        click.echo(
            "Current Global Loss: {0:.2f} ({1})".format(
                float(val["value"]), val["date"]
            )
        )
    else:
        click.echo("No Validation Loss Data yet")
    if prec.status_code < 300 and "val_global_Prec@1 @ 0" in prec.json():
        val = prec.json()["val_global_Prec@1 @ 0"][0]
        click.echo(
            "Current Global Precision: {0:.2f} ({1})".format(
                float(val["value"]), val["date"]
            )
        )
    else:
        click.echo("No Validation Precision Data yet")


@cli_group.command()
@click.argument("folder", type=click.Path(exists=False, file_okay=False, dir_okay=True))
@click.option(
    "--filter", "-f", default=None, type=str, help="String filter to filter runs"
)
@click.option("--dashboard-url", "-u", default=None, type=str)
def charts(folder, filter, dashboard_url):
    """Chart the results of benchmark runs

    Save generated charts in FOLDER
    """
    folder = Path(folder)
    if not folder.exists():
        folder.mkdir(parents=True)
    loaded = setup_client_from_config()

    client = ApiClient(in_cluster=False, url=dashboard_url, load_config=not loaded)

    ret = client.get_runs()
    runs = ret.result().json()
    runs = [r for r in runs if r["state"] == "finished"]

    if filter:
        runs = [r for r in runs if filter in r["name"]]

    options = {i: r for i, r in enumerate(runs, start=0)}

    if len(options) < 2:
        click.echo("At least two finished runs are needed to create a summary")
        return

    options["all"] = {"name": "*all runs*"}

    prompt = 'Select the runs to generate a summary for (e.g. "0 1 2"): \n\t{}'.format(
        "\n\t".join("{} [{}]".format(r["name"], i) for i, r in options.items())
    )

    choice = click.prompt(
        prompt,
        default=0,
        type=click.Choice([options.keys()]),
        show_choices=False,
        value_proc=lambda x: runs
        if "all" in x
        else [options[int(i)] for i in x.split(" ")],
    )

    if len(choice) < 2:
        click.echo("At least two finished runs are needed to create a summary")
        return

    results = []

    def _get_metric(name, run):
        """Gets a metric from the dashboard."""
        name = "global_cum_{} @ 0".format(name)
        return float(
            client.get_run_metrics(run["id"], metric_filter=name, last_n=1)
            .result()
            .json()[name][0]["value"]
        )

    for run in choice:
        agg = _get_metric("agg", run)

        backprop = _get_metric("backprop", run)

        batch_load = _get_metric("batch_load", run)

        comp_loss = _get_metric("comp_loss", run)

        comp_metrics = _get_metric("comp_metrics", run)

        fwd_pass = _get_metric("fwd_pass", run)

        opt_step = _get_metric("opt_step", run)

        compute = (
            fwd_pass
            + comp_loss
            + backprop
            + opt_step
            + (agg if run["num_workers"] == 1 else 0)
        )
        communicate = agg if run["num_workers"] != 1 else 0

        results.append(
            (
                run["name"],
                compute,
                communicate,
                comp_metrics,
                batch_load,
                str(run["num_workers"]),
            )
        )

    results = sorted(results, key=lambda x: x[5])
    names, compute, communicate, metrics, batch_load, num_workers = zip(*results)

    width = 0.35
    fig, ax = plt.subplots()

    ax.bar(num_workers, compute, width, label="Compute")
    ax.bar(num_workers, communicate, width, label="Communication")

    ax.set_ylabel("Time (s)")
    ax.set_title("Total time by number of workers")
    ax.legend()
    plt.savefig(folder / "total_time.png", dpi=150)

    fig, ax = plt.subplots()

    combined = [c + r for _, c, r, _, _, _ in results]

    speedup = [combined[0] / c for c in combined]

    ax.bar(num_workers, speedup, width)

    ax.set_ylabel("Speedup factor")
    ax.set_title("Speedup")
    plt.savefig(folder / "speedup.png", dpi=150)

    fig, ax = plt.subplots()

    ax.bar(num_workers, compute, width, label="Compute")
    ax.bar(num_workers, communicate, width, label="Communication")
    ax.bar(num_workers, metrics, width, label="Metrics Computation")
    ax.bar(num_workers, batch_load, width, label="Batch Loading")

    ax.set_ylabel("Time (s)")
    ax.set_title("Total time by number of workers")
    ax.legend()
    plt.savefig(folder / "time_for_all_phases.png", dpi=150)

    click.echo("Summary created in {}".format(folder))


@cli_group.command()
def get_dashboard_url():
    """Returns the dashboard URL of the current cluster"""
    loaded = setup_client_from_config()

    if not loaded:
        click.echo("No Cluster config found")
        return

    client = ApiClient(in_cluster=False, load_config=False)

    click.echo(client.endpoint.replace("api/", ""))


@cli_group.command()
@click.argument("name", type=str)
@click.option("--dashboard-url", "-u", default=None, type=str)
def delete(name, dashboard_url):
    """Delete a benchmark run"""
    loaded = setup_client_from_config()

    client = ApiClient(in_cluster=False, url=dashboard_url, load_config=not loaded)

    ret = client.get_runs()
    runs = ret.result().json()

    try:
        run = next(r for r in runs if r["name"] == name)
    except StopIteration:
        click.echo("Run not found")
        return

    del run["job_id"]
    del run["job_metadata"]

    client.delete_run(run["id"])


@cli_group.command()
@click.argument("name", type=str)
@click.option("--output", "-o", default="results.zip", type=str)
@click.option("--dashboard-url", "-u", default=None, type=str)
def download(name, output, dashboard_url):
    """Download the results of a benchmark run"""
    loaded = setup_client_from_config()

    client = ApiClient(in_cluster=False, url=dashboard_url, load_config=not loaded)

    ret = client.get_runs()
    runs = ret.result().json()

    run = next(r for r in runs if r["name"] == name)

    ret = client.download_run_metrics(run["id"])

    with open(output, "wb") as f:
        f.write(ret.result().content)


@cli_group.command("list-clusters")
def list_clusters():
    """List all currently configured clusters."""
    config = get_config()

    sections = [
        s.split(".", 1)
        for s in config.sections()
        if s.startswith("gke.") or s.startswith("aws.")
    ]

    if len(sections) == 0:
        click.echo("No clusters are currently configured")
        return

    current_provider = config.get("general", "provider", fallback=None)
    current_cluster = None

    if current_provider == "gke":
        current_cluster = config.get("gke", "current-cluster", fallback=None)
    elif current_provider == "aws":
        current_cluster = config.get("aws", "current-cluster", fallback=None)
    elif current_provider == "kind":
        current_cluster = config.get("kind", "current-cluster", fallback=None)

    gke = [
        name + " *" if current_provider == "gke" and current_cluster == name else name
        for t, name in sections
        if t == "gke"
    ]
    aws = [
        name + " *" if current_provider == "aws" and current_cluster == name else name
        for t, name in sections
        if t == "aws"
    ]
    kind = [
        name + " *" if current_provider == "kind" and current_cluster == name else name
        for t, name in sections
        if t == "kind"
    ]

    message = "Clusters:"

    if gke:
        message += "\n\tGoogle Cloud:\n\t\t{}".format("\n\t\t".join(gke))
    if aws:
        message += "\n\tAWS:\n\t\t{}".format("\n\t\t".join(aws))
    if kind:
        message += "\n\Kind:\n\t\t{}".format("\n\t\t".join(kind))

    click.echo(message)


@cli_group.group("set-cluster")
def set_cluster():
    """Set the current cluster to use."""
    pass


@set_cluster.command("gcloud")
@click.argument("name", type=str)
def set_gcloud_cluster(name):
    """Set current cluster to a gcloud cluster."""
    config = get_config()

    if not config.has_section("gke.{}".format(name)):
        click.echo("Cluster {} not found".format(name))

    config.set("general", "provider", "gke")
    config.set("gke", "current-cluster", name)
    write_config(config)

    click.echo("Ok")


@set_cluster.command("aws")
@click.argument("name", type=str)
def set_aws_cluster(name):
    """Set current cluster to an aws cluster."""
    config = get_config()

    if not config.has_section("aws.{}".format(name)):
        click.echo("Cluster {} not found".format(name))

    config.set("general", "provider", "aws")
    config.set("aws", "current-cluster", name)
    write_config(config)

    click.echo("Ok")


@set_cluster.command("kind")
@click.argument("name", type=str)
def set_kind_cluster(name):
    """Set current cluster to an aws cluster."""
    config = get_config()

    if not config.has_section("kind.{}".format(name)):
        click.echo("Cluster {} not found".format(name))

    config.set("general", "provider", "kind")
    config.set("kind", "current-cluster", name)
    write_config(config)

    click.echo("Ok")


@cli_group.group("delete-cluster")
def delete_cluster():
    """Delete a cluster."""
    pass


@delete_cluster.command("gcloud")
@click.argument("name", type=str)
@click.option("--project", "-p", default=None, type=str)
@click.option("--zone", "-z", default="europe-west1-b", type=str)
def delete_gcloud(name, zone, project):
    gcloud_delete_cluster(name, zone, project)
    delete_gcloud_cluster(name)


@delete_cluster.command("kind")
@click.argument("name", type=str)
def delete_kind(name):
    # Get registry name
    config = get_config()
    registry_name = config.get("kind.{}".format(name), "registry_name")

    # Delete cluster
    kind_delete_cluster(name, registry_name)

    # Delete cluster from config
    delete_kind_cluster(name)

    click.echo("Cluster deleted.")


@delete_cluster.command("aws")
@click.argument("name", type=str)
def delete_aws(name):
    sts = boto3.client("sts")
    try:
        sts.get_caller_identity()
    except botocore.exceptions.ClientError:
        raise click.UsageError(
            "Couldn't find aws credentials. Install the aws"
            " sdk ( https://docs.aws.amazon.com/cli/latest/userguide/install-cliv2-linux.html ) and "
            "run 'aws configure' to login and create "
            "your credentials."
        )

    # delete nodegroup
    stackName = "eks-auto-scaling-group-" + name

    cloudFormation = boto3.client("cloudformation")
    cloudFormation.delete_stack(StackName=stackName)

    waiter = cloudFormation.get_waiter("stack_delete_complete")
    click.echo("Waiting for nodegroup to be deleted.")
    waiter.wait(StackName=stackName)

    # delete EKS cluster
    eks = boto3.client("eks")
    eks.delete_cluster(name=name)
    waiter = eks.get_waiter("cluster_deleted")
    click.echo("Waiting for cluster to be deleted. This can take up to ten minutes.")
    waiter.wait(name=name)

    # delete VPC stack
    stack_name = name + "-stack"
    cf_client = boto3.client("cloudformation")
    cf_client.delete_stack(StackName=stack_name)
    waiter = cf_client.get_waiter("stack_delete_complete")
    click.echo("Waiting for the VPC stack to be deleted.")
    waiter.wait(StackName=stack_name)

    delete_aws_cluster(name)

    click.echo("Cluster deleted.")


@cli_group.group("create-cluster")
def create_cluster():
    """Create a new cluster."""
    pass


@create_cluster.command("gcloud")
@click.argument("num_workers", type=int, metavar="num-workers")
@click.argument("release", type=str)
@click.option("--kubernetes-version", "-k", type=str, default="1.15")
@click.option("--machine-type", "-t", default="n1-standard-4", type=str)
@click.option("--disk-size", "-d", default=50, type=int)
@click.option("--num-cpus", "-c", default=4, type=int)
@click.option("--num-gpus", "-g", default=0, type=int)
@click.option("--gpu-type", default="nvidia-tesla-k80", type=str)
@click.option("--zone", "-z", default="europe-west1-b", type=str)
@click.option("--project", "-p", default=None, type=str)
@click.option("--preemptible", "-e", is_flag=True)
@click.option("--custom-value", "-v", multiple=True)
@click.option("--chart-location", "-cl", default=None, type=str)
def create_gcloud(
    num_workers,
    release,
    kubernetes_version,
    machine_type,
    disk_size,
    num_cpus,
    num_gpus,
    gpu_type,
    zone,
    project,
    preemptible,
    custom_value,
    chart_location,
):
    import google.auth
    from google.auth.exceptions import DefaultCredentialsError

    try:
        credentials, default_project = google.auth.default()
    except DefaultCredentialsError:
        raise click.UsageError(
            "Couldn't find gcloud credentials. Install the gcloud"
            " sdk ( https://cloud.google.com/sdk/docs/quickstart-linux ) and "
            "run 'gcloud auth application-default login' to login and create "
            "your credentials."
        )

    if not project:
        project = default_project

    name = "{}-{}".format(release, num_workers)
    name_path = "projects/{}/locations/{}/clusters/".format(project, zone)

    click.echo("Creating Cluster")

    gclient, fw_name, firewalls = gcloud_create_cluster(
        name=name,
        name_path=name_path,
        num_workers=num_workers,
        num_gpus=num_gpus,
        gpu_type=gpu_type,
        machine_type=machine_type,
        disk_size=disk_size,
        preemptible=preemptible,
        kubernetes_version=kubernetes_version,
        project=project,
    )

    cluster = gclient.get_cluster(None, None, None, name=os.path.join(name_path, name))
    kube_context = setup_gcloud_kube_client(
        cluster.endpoint, cluster.name, cluster.zone, project
    )

    if num_gpus > 0:
        deploy_nvidia_daemonset()

    custom_chart = {
        "name": "mlbench-helm",
        "source": {
            "type": "git" if chart_location is None else "directory",
            "location": "https://github.com/mlbench/mlbench-helm"
            if chart_location is None
            else chart_location,
        },
    }

    click.echo("Deploying chart")
    deploy_chart(
        num_workers=num_workers - 1,
        num_gpus=num_gpus,
        num_cpus=num_cpus - 1,
        release_name=name,
        custom_value=custom_value,
        custom_chart=custom_chart,
        kube_context=kube_context,
    )

    # open port in firewall
    mlbench_client = ApiClient(in_cluster=False, load_config=False)
    firewall_body = {
        "name": fw_name,
        "direction": "INGRESS",
        "sourceRanges": "0.0.0.0/0",
        "allowed": [{"IPProtocol": "tcp", "ports": [mlbench_client.port]}],
    }

    firewalls.insert(project=project, body=firewall_body).execute()

    add_gcloud_cluster(name, cluster, project)

    click.echo("MLBench successfully deployed")


@create_cluster.command("aws")
@click.argument("num_workers", type=int, metavar="num-workers")
@click.argument("release", type=str)
@click.option("--kubernetes-version", "-k", type=str, default="1.15")
@click.option("--machine-type", "-t", default="t2.medium", type=str)
@click.option("--num-cpus", "-c", default=1, type=int)
@click.option("--num-gpus", "-g", default=0, type=int)
@click.option("--custom-value", "-v", multiple=True)
@click.option("--ami-id", "-a", default="ami-0ef76ba092ce4e253", type=str)
@click.option("--ssh-key", "-a", default="eksNodeKey", type=str)
def create_aws(
    num_workers,
    release,
    kubernetes_version,
    machine_type,
    num_cpus,
    num_gpus,
    custom_value,
    ami_id,
    ssh_key,
):
    sts = boto3.client("sts")
    try:
        sts.get_caller_identity()
    except botocore.exceptions.ClientError:
        raise click.UsageError(
            "Couldn't find aws credentials. Install the aws"
            " sdk ( https://docs.aws.amazon.com/cli/latest/userguide/install-cliv2-linux.html ) and "
            "run 'aws configure' to login and create "
            "your credentials."
        )
    name = "{}-{}".format(release, num_workers)
    nodeGroupName = name + "-node-group"

    kube_context, cf_client, stackName, cluster = aws_create_cluster(
        name,
        nodeGroupName,
        num_workers,
        machine_type,
        ssh_key,
        ami_id,
        kubernetes_version,
    )
    kube_config.load_kube_config(context=kube_context)

    if num_gpus > 0:
        deploy_nvidia_daemonset_aws()

    deploy_chart(
        num_workers=num_workers - 1,
        num_gpus=num_gpus,
        num_cpus=num_cpus - 1,
        release_name=name,
        custom_value=custom_value,
        kube_context=kube_context,
    )

    # open port in firewall
    mlbench_client = ApiClient(in_cluster=False, load_config=False)
    mlbench_port = mlbench_client.port
    r = cf_client.describe_stack_resources(
        StackName=stackName, LogicalResourceId="NodeSecurityGroup"
    )
    secGroupId = r["StackResources"][0]["PhysicalResourceId"]

    ec2 = boto3.client("ec2")
    ec2.authorize_security_group_ingress(
        GroupId=secGroupId,
        IpPermissions=[
            {
                "FromPort": mlbench_port,
                "IpProtocol": "tcp",
                "IpRanges": [
                    {
                        "CidrIp": "0.0.0.0/0",
                    },
                ],
                "ToPort": mlbench_port,
            },
        ],
    )
    add_aws_cluster(name, cluster)
    click.echo("MLBench successfully deployed")


@create_cluster.command("kind")
@click.argument("num_workers", type=int, metavar="num-workers")
@click.argument("release", type=str)
@click.option("--chart-location", "-cl", default=None, type=str)
@click.option("--registry_name", "-r", default="kind-registry", type=str)
@click.option("--registry_port", "-p", default="5000", type=str)
@click.option("--host_port", "-h", default="5000", type=str)
@click.option("--num-cpus", "-c", default=1, type=int)
@click.option("--num-gpus", "-g", default=0, type=int)
@click.option("--custom-value", "-v", multiple=True)
@click.option("--kubernetes-version", "-k", type=str, default="1.15")
def create_kind(
    num_workers,
    release,
    chart_location,
    registry_name,
    registry_port,
    host_port,
    num_cpus,
    num_gpus,
    custom_value,
    kubernetes_version,
):
    name = "{}-{}".format(release, num_workers)

    create_local_registry(registry_name, registry_port, host_port)

    kind_create_cluster(
        name=name,
        num_workers=num_workers,
        registry_name=registry_name,
        registry_port=registry_port,
        kubernetes_version=kubernetes_version,
    )

    kube_context = "kind-{}".format(name)
    kube_config.load_kube_config(context=kube_context)

    custom_chart = {
        "name": "mlbench-helm",
        "source": {
            "type": "git" if chart_location is None else "directory",
            "location": "https://github.com/mlbench/mlbench-helm"
            if chart_location is None
            else chart_location,
        },
    }
    click.echo("Deploying chart")
    deploy_chart(
        num_workers=num_workers - 1,
        num_gpus=num_gpus,
        num_cpus=num_cpus,
        release_name=name,
        custom_value=custom_value,
        custom_chart=custom_chart,
        kube_context=kube_context,
    )

    add_kind_cluster(name, registry_name)

    click.echo("MLBench successfully deployed")


def get_config_path():
    """Gets the path to the user config."""
    user_dir = user_data_dir("mlbench", "mlbench")
    return os.path.join(user_dir, "mlbench.ini")


def get_config():
    """Get the current users' config."""
    path = get_config_path()

    config = configparser.ConfigParser()

    if os.path.exists(path):
        config.read(path)

    if not config.has_section("general"):
        config.add_section("general")

    if not config.has_section("gke"):
        config.add_section("gke")

    if not config.has_section("aws"):
        config.add_section("aws")

    if not config.has_section("kind"):
        config.add_section("kind")

    return config


def write_config(config):
    """Save a config object."""
    path = get_config_path()

    if not os.path.exists(path):
        os.makedirs(os.path.dirname(path))

    with open(path, "w") as configfile:
        config.write(configfile)


def add_gcloud_cluster(name, cluster, project):
    """Add a gcloud cluster to config."""
    config = get_config()

    config.set("general", "provider", "gke")
    config.set("gke", "current-cluster", name)

    section = "gke.{}".format(name)

    if not config.has_section(section):
        config.add_section(section)
    config.set(section, "endpoint", cluster.endpoint)
    config.set(section, "name", cluster.name)
    config.set(section, "zone", cluster.zone)
    config.set(section, "project", project)

    write_config(config)


def delete_gcloud_cluster(name):
    """Delete a gcloud cluster from config."""
    config = get_config()
    config.remove_section("gke.{}".format(name))

    if config.get("gke", "current-cluster", fallback=None):
        config.set("gke", "current-cluster", "")

    write_config(config)


def add_aws_cluster(name, cluster):
    """Add an aws cluster to config."""
    config = get_config()

    config.set("general", "provider", "aws")
    config.set("aws", "current-cluster", name)

    section = "aws.{}".format(name)

    if not config.has_section(section):
        config.add_section(section)

    config.set(section, "cluster", cluster["cluster"]["endpoint"])

    write_config(config)


def delete_aws_cluster(name):
    """Delete an aws cluster from config."""
    config = get_config()
    config.remove_section("aws.{}".format(name))

    if config.get("aws", "current-cluster", fallback=None):
        config.set("aws", "current-cluster", "")

    write_config(config)


def add_kind_cluster(name, registry_name):
    """Add an kind cluster to config."""
    config = get_config()

    config.set("general", "provider", "kind")
    config.set("kind", "current-cluster", name)

    section = "kind.{}".format(name)

    if not config.has_section(section):
        config.add_section(section)

    config.set(section, "cluster", name)
    config.set(section, "registry_name", registry_name)

    write_config(config)


def delete_kind_cluster(name):
    """Delete a kind cluster from config."""
    config = get_config()
    config.remove_section("kind.{}".format(name))

    if config.get("kind", "current-cluster", fallback=None):
        config.set("kind", "current-cluster", "")

    write_config(config)


def setup_client_from_config():
    """Setup the current kubernetes config."""
    config = get_config()

    provider = config.get("general", "provider", fallback=None)

    if not provider:
        return False

    if provider == "gke":
        return setup_gke_client_from_config(config)
    if provider == "kind":
        return setup_kind_client_from_config(config)
    if provider == "aws":
        return setup_aws_client_from_config(config)

    else:
        raise NotImplementedError()


def setup_kind_client_from_config(config):
    """Setup a kubernrtes cluster for kind from current config."""

    cluster = config.get("kind", "current-cluster", fallback=None)
    if not cluster:
        raise click.UsageError(
            "No kind cluster selected, create a new one with `mlbench create-cluster`"
            " or set one with `mlbench set-cluster`"
        )

    cluster = config.get("kind.{}".format(cluster), "cluster", fallback=None)
    if not cluster:
        return False

    kube_config.load_kube_config(context="kind-{}".format(cluster))
    return True


def setup_gke_client_from_config(config):
    """Setup a kubernetes cluster for gke from current config."""
    cluster = config.get("gke", "current-cluster", fallback=None)
    if not cluster:
        raise click.UsageError(
            "No gcloud cluster selected, create a new one with `mlbench create-cluster`"
            " or set one with `mlbench set-cluster`"
        )

    section = "gke.{}".format(cluster)
    cluster_endpoint = config.get(section, "endpoint", fallback=None)
    if not cluster:
        return False

    cluster_name = config.get(section, "name", fallback=None)
    if not cluster_name:
        return False

    cluster_zone = config.get(section, "zone", fallback=None)
    if not cluster_zone:
        return False

    project = config.get(section, "project", fallback=None)
    if not project:
        return False

    kube_context = setup_gcloud_kube_client(
        cluster_endpoint, cluster_name, cluster_zone, project
    )
    kube_config.load_kube_config(context=kube_context)
    return True


def setup_aws_client_from_config(config):
    """Setup a kubernetes cluster for aws from current config."""
    name = config.get("aws", "current-cluster", fallback=None)
    if not name:
        raise click.UsageError(
            "No gcloud cluster selected, create a new one with `mlbench create-cluster`"
            " or set one with `mlbench set-cluster`"
        )

    kube_context = setup_aws_kube_client(name)
    kube_config.load_kube_config(context=kube_context)
    return True


if __name__ == "__main__":
    sys.exit(cli_group())  # pragma: no cover



from torch import nn

from mlbench_core.models.pytorch.transformer.decoder import TransformerDecoder
from mlbench_core.models.pytorch.transformer.encoder import TransformerEncoder
from mlbench_core.models.pytorch.transformer.modules import build_embedding

DEFAULT_MAX_SOURCE_POSITIONS = 256
DEFAULT_MAX_TARGET_POSITIONS = 256


class TransformerModel(nn.Module):
    """Transformer model

    This model uses MultiHeadAttention as described in
    :cite:`NIPS2017_7181`

    Args:
        args: Arguments of model. All arguments should be accessible via `__getattribute__` method
        src_dict (:obj:`mlbench_core.dataset.nlp.pytorch.wmt17.Dictionary`): Source dictionary
        trg_dict (:obj:`mlbench_core.dataset.nlp.pytorch.wmt17.Dictionary`): Target dictionary
    """

    def __init__(self, args, src_dict, trg_dict):
        super().__init__()
        self._is_generation_fast = False
        if not hasattr(args, "max_source_positions"):
            args.max_source_positions = DEFAULT_MAX_SOURCE_POSITIONS
        if not hasattr(args, "max_target_positions"):
            args.max_target_positions = DEFAULT_MAX_TARGET_POSITIONS

        # Define embedding layer
        if args.share_all_embeddings:
            if src_dict != trg_dict:
                raise ValueError("share_all_embeddings requires a joined dictionary")
            if args.encoder_embed_dim != args.decoder_embed_dim:
                raise ValueError(
                    "share_all_embeddings requires encoder_embed_dim to match decoder_embed_dim"
                )
            if args.decoder_embed_path and (
                args.decoder_embed_path != args.encoder_embed_path
            ):
                raise ValueError(
                    "share_all_embeddings not compatible with decoder_embed_path"
                )
            encoder_embed_tokens = build_embedding(
                src_dict, args.encoder_embed_dim, args.encoder_embed_path
            )
            decoder_embed_tokens = encoder_embed_tokens
            args.share_decoder_input_output_embed = True
        else:
            encoder_embed_tokens = build_embedding(
                src_dict, args.encoder_embed_dim, args.encoder_embed_path
            )
            decoder_embed_tokens = build_embedding(
                trg_dict, args.decoder_embed_dim, args.decoder_embed_path
            )
        self.encoder = TransformerEncoder(args, src_dict, encoder_embed_tokens)
        self.decoder = TransformerDecoder(args, trg_dict, decoder_embed_tokens)

    def forward(
        self,
        src_tokens,
        src_lengths,
        prev_output_tokens,
    ):
        """
        Run the forward pass of the transformer model.

        Args:
            src_tokens (:obj:`torch.Tensor`): Source tokens
            src_lengths (:obj:`torch.Tensor`): Source sentence lengths
            prev_output_tokens (:obj:`torch.Tensor`): Previous output tokens

        Returns:
            (:obj:`torch.Tensor`, Optional[:obj:`torch.Tensor`]):
                The model output, and attention weights if needed
        """
        encoder_out = self.encoder(src_tokens, src_lengths=src_lengths)
        decoder_out = self.decoder(prev_output_tokens, encoder_out=encoder_out)
        return decoder_out

    def max_positions(self):
        """Maximum length supported by the model."""
        return self.encoder.max_positions(), self.decoder.max_positions()

    def max_decoder_positions(self):
        """Maximum length supported by the decoder.

        Returns:
            (int)
        """
        return self.decoder.max_positions()



import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter

from mlbench_core.models.pytorch.gnmt import attn_score


class AttentionScore(torch.autograd.Function):
    @staticmethod
    def forward(ctx, att_query, att_keys, bias, linear_att):
        score = attn_score.forward(att_query, att_keys, bias, linear_att)
        ctx.save_for_backward(att_query, att_keys, bias, linear_att)
        return score

    @staticmethod
    def backward(ctx, grad_output):
        att_query, att_keys, bias, linear_att = ctx.saved_tensors
        grad_query, grad_keys, grad_bias, grad_linear_att = attn_score.backward(
            grad_output, att_query, att_keys, bias, linear_att
        )
        return grad_query, grad_keys, grad_bias, grad_linear_att


fused_calc_score = AttentionScore.apply


class BahdanauAttention(nn.Module):
    """
    Bahdanau Attention (https://arxiv.org/abs/1409.0473)
    Implementation is very similar to tf.contrib.seq2seq.BahdanauAttention

    Args:
        query_size (int): feature dimension for query
        key_size (int): feature dimension for keys
        num_units (int): internal feature dimension
        normalize (bool): whether to normalize energy term.
            Default: `False`
        init_weight (float): range for uniform initializer used to initialize
            Linear key and query transform layers and linear_att vector.
            Default: 0.1
    """

    def __init__(
        self,
        query_size,
        key_size,
        num_units,
        normalize=False,
        init_weight=0.1,
        fusion=True,
    ):
        super(BahdanauAttention, self).__init__()

        self.normalize = normalize
        self.num_units = num_units

        self.linear_q = nn.Linear(query_size, num_units, bias=False)
        self.linear_k = nn.Linear(key_size, num_units, bias=False)
        nn.init.uniform_(self.linear_q.weight.data, -init_weight, init_weight)
        nn.init.uniform_(self.linear_k.weight.data, -init_weight, init_weight)

        self.linear_att = Parameter(torch.Tensor(num_units))

        self.mask = None

        if self.normalize:
            self.normalize_scalar = Parameter(torch.Tensor(1))
            self.normalize_bias = Parameter(torch.Tensor(num_units))
        else:
            self.register_parameter("normalize_scalar", None)
            self.register_parameter("normalize_bias", None)

        self.fusion = fusion
        self.reset_parameters(init_weight)

    def reset_parameters(self, init_weight):
        """
        Sets initial random values for trainable parameters.

        Args:
            init_weight (float):
        """
        stdv = 1.0 / math.sqrt(self.num_units)
        self.linear_att.data.uniform_(-init_weight, init_weight)

        if self.normalize:
            self.normalize_scalar.data.fill_(stdv)
            self.normalize_bias.data.zero_()

    def set_mask(self, context_len, context):
        """
        Sets self.mask which is applied before softmax
        ones for inactive context fields, zeros for active context fields

        Args:
            context_len (`obj`:torch.Tensor):
            context (`obj`:torch.Tensor): (t_k x b x n)

        Returns:

        """
        max_len = context.size(0)

        indices = torch.arange(0, max_len, dtype=torch.int64, device=context.device)
        self.mask = indices >= (context_len.unsqueeze(1))

    def calc_score(self, att_query, att_keys):
        """
        Calculate Bahdanau score

        Args:
            att_query (`obj`:torch.Tensor):
            att_keys (`obj`:torch.Tensor):

        Returns:

        """

        b, t_k, n = att_keys.size()
        t_q = att_query.size(1)

        att_query = att_query.unsqueeze(2).expand(b, t_q, t_k, n)
        att_keys = att_keys.unsqueeze(1).expand(b, t_q, t_k, n)
        sum_qk = att_query + att_keys

        if self.normalize:
            sum_qk = sum_qk + self.normalize_bias
            linear_att = self.linear_att / self.linear_att.norm()
            linear_att = linear_att * self.normalize_scalar
        else:
            linear_att = self.linear_att

        out = torch.tanh(sum_qk).matmul(linear_att)
        return out

    def forward(self, query, keys):
        """

        Args:
            query (`obj`:torch.Tensor): (t_q x b x n)
            keys (`obj`:torch.Tensor): (t_k x b x n)

        Returns:
            (context, scores_normalized)
        context: (t_q x b x n)
        scores_normalized: (t_q x b x t_k)

        """

        # first dim of keys and query has to be 'batch', it's needed for bmm
        keys = keys.transpose(0, 1)
        if query.dim() == 3:
            query = query.transpose(0, 1)

        if query.dim() == 2:
            single_query = True
            query = query.unsqueeze(1)
        else:
            single_query = False

        b = query.size(0)
        t_k = keys.size(1)
        t_q = query.size(1)

        # FC layers to transform query and key
        processed_query = self.linear_q(query)
        processed_key = self.linear_k(keys)

        # scores: (b x t_q x t_k)
        if self.fusion:
            linear_att = self.linear_att / self.linear_att.norm()
            linear_att = linear_att * self.normalize_scalar
            scores = fused_calc_score(
                processed_query, processed_key, self.normalize_bias, linear_att
            )
        else:
            scores = self.calc_score(processed_query, processed_key)

        if self.mask is not None:
            mask = self.mask.unsqueeze(1).expand(b, t_q, t_k)
            # I can't use -INF because of overflow check in pytorch
            scores.data.masked_fill_(mask, -65504.0)

        # Normalize the scores, softmax over t_k
        scores_normalized = F.softmax(scores, dim=-1)

        # Calculate the weighted average of the attention inputs according to
        # the scores
        # context: (b x t_q x n)
        context = torch.bmm(scores_normalized, keys)

        if single_query:
            context = context.squeeze(1)
            scores_normalized = scores_normalized.squeeze(1)
        else:
            context = context.transpose(0, 1)
            scores_normalized = scores_normalized.transpose(0, 1)

        return context, scores_normalized



import torch
from mosestokenizer import MosesDetokenizer

import mlbench_core.dataset.nlp.pytorch.wmt16.wmt16_config as wmt16_config
from mlbench_core.utils.pytorch.inference.beam_search import SequenceGenerator


class Translator:
    """
    Translator that output translated sentences from GNMT model by using a Sequence Generator

    Args:
        model (`obj`:torch.nn.Module): Model to use
        trg_tokenizer (`obj`:mlbench_core.dataset.nlp.pytorch.wmt16.WMT16Tokenizer): The target tokenizer
    """

    def __init__(
        self,
        model,
        trg_tokenizer,
        trg_lang="de",
        beam_size=5,
        len_norm_factor=0.6,
        len_norm_const=5.0,
        cov_penalty_factor=0.1,
        max_seq_len=150,
    ):

        self.model = model
        self.tokenizer = trg_tokenizer
        self.insert_target_start = [wmt16_config.BOS]
        self.insert_src_start = [wmt16_config.BOS]
        self.insert_src_end = [wmt16_config.EOS]
        self.beam_size = beam_size
        self.trg_lang = trg_lang

        self.generator = SequenceGenerator(
            model=self.model,
            beam_size=beam_size,
            max_seq_len=max_seq_len,
            len_norm_factor=len_norm_factor,
            len_norm_const=len_norm_const,
            cov_penalty_factor=cov_penalty_factor,
        )

    def get_detokenized_target(self, trg, batch_size):
        targets = []
        with MosesDetokenizer(self.trg_lang) as detok:
            for i in range(batch_size):
                t = self.tokenizer.detokenize(trg[:, i].tolist())
                t = detok(t.split())
                targets.append(t)

        return targets

    def translate(self, src, trg):
        """Given a source a target tokenized tensors, outputs the
        non-tokenized translation from the model, as well as the non-tokenized target

        Args:
            src:
            trg:

        Returns:

        """
        src, src_len = src
        trg, trg_len = trg
        device = next(self.model.parameters()).device

        batch_size = src.shape[1]

        bos = [self.insert_target_start] * (batch_size * self.beam_size)
        bos = torch.tensor(bos, dtype=torch.int64, device=device).view(1, -1)

        if self.beam_size == 1:
            generator = self.generator.greedy_search
        else:
            generator = self.generator.beam_search

        with torch.no_grad():
            context = self.model.encode(src, src_len)
            context = [context, src_len, None]
            preds, lengths, counter = generator(batch_size, bos, context)

        preds = preds.cpu()
        targets = self.get_detokenized_target(trg, batch_size)

        output = []
        with MosesDetokenizer(self.trg_lang) as detokenizer:
            for pred in preds:
                pred = pred.tolist()
                detok = self.tokenizer.detokenize(pred)
                detok = detokenizer(detok.split())
                output.append(detok)

        return output, targets



from torch.optim import SGD
from torch.optim.optimizer import required

from mlbench_core.aggregation.pytorch.decentralized import DecentralizedAggregation


class DecentralizedSGD(SGD):
    r"""Implements decentralized stochastic gradient descent (optionally with momentum).

    Args:
        rank (int): rank of current process in the network
        neighbors (list): list of ranks of the neighbors of current process
        model (:obj:`nn.Module`): model which contains parameters for SGD
        lr (float): learning rate
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)
        average_world (bool): Whether to average models on the world_size (default: `True`)
        use_cuda (bool): Whether to use cuda tensors for aggregation
        by_layer (bool): Aggregate by layer instead of all layers at once
    """

    def __init__(
        self,
        rank=None,
        neighbors=None,
        model=None,
        lr=required,
        momentum=0,
        dampening=0,
        weight_decay=0,
        nesterov=False,
        average_world=True,
        use_cuda=False,
        by_layer=False,
    ):
        if not rank:
            raise ValueError('"rank" not set for optimizer')
        if not neighbors:
            raise ValueError('"neighbors" not set for optimizer')
        if not model:
            raise ValueError('"model" not set for optimizer')
        super(DecentralizedSGD, self).__init__(
            model.parameters(), lr, momentum, dampening, weight_decay, nesterov
        )

        if average_world:
            self.agg_mode = "avg_world"
        else:
            raise NotImplementedError("Only average model is supported right now.")

        self.model = model
        self.agg = DecentralizedAggregation(
            rank, neighbors, use_cuda=use_cuda
        ).agg_model(by_layer=by_layer)

    def step(self, closure=None, tracker=None):
        """Aggregates the gradients and performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
            tracker (:obj:`mlbench_core.utils.Tracker`, optional) The current tracker
        """
        loss = super(DecentralizedSGD, self).step(closure=closure)
        if tracker:
            tracker.record_batch_opt_step()
        # Averaging the model after updating the gradient separately.
        self.agg(self.model, self.agg_mode)
        if tracker:
            tracker.record_batch_agg()
        return loss



from .embeddings import PositionalEmbedding, build_embedding
from .layers import TransformerDecoderLayer, TransformerEncoderLayer

__ALL__ = [
    PositionalEmbedding,
    build_embedding,
    TransformerDecoderLayer,
    TransformerEncoderLayer,
]



from abc import ABC

import torch
import torch.distributed as dist
from torch.nn.utils import clip_grad_norm_
from torch.optim import SGD, Adam
from torch.optim.optimizer import required

from mlbench_core.aggregation.pytorch.centralized import (
    AVG_CUSTOM,
    AVG_WORLD,
    AllReduceAggregation,
    PowerAggregation,
)
from mlbench_core.optim.pytorch.optim import SparsifiedSGD


class GenericCentralizedOptimizer(ABC):
    """Implements a generic centralized (synchronous) optimizer with `AllReduceAggregation`.
    Averages the reduced parameters over the world size, after aggregation.
    Can aggregate gradients or weights, by layers or all at once.

    Args:
        world_size (int): Size of the network
        model (:obj:`nn.Module`): Model which contains parameters for SGD
        use_cuda (bool): Whether to use cuda tensors for aggregation
        by_layer (bool): Aggregate by layer instead of all layers at once
        agg_grad (bool): Aggregate the gradients before updating weights. If `False`,
            weights will be updated and then reduced across all workers. (default: `True`)
    """

    def __init__(
        self,
        world_size,
        model,
        use_cuda=False,
        by_layer=False,
        divide_before=False,
        agg_grad=True,
    ):
        if not world_size:
            raise ValueError('"world_size" not set for optimizer')
        if not model:
            raise ValueError('"model" not set for optimizer')

        self.agg_mode = AVG_WORLD
        self.model = model
        self.agg_grad = agg_grad
        agg = AllReduceAggregation(
            world_size=world_size, use_cuda=use_cuda, divide_before=divide_before
        )  # Divide after reduction
        if agg_grad:
            self.agg = agg.agg_grad(by_layer=by_layer)
        else:
            self.agg = agg.agg_model(by_layer=by_layer)
            self.agg(self.model, self.agg_mode)  # Agg params once at init

        self.optimizer = None

    def step(self, closure=None, tracker=None):
        """Aggregates the gradients and performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
            tracker (:obj:`mlbench_core.utils.Tracker`, optional) The current tracker
        """
        if self.agg_grad:
            self.agg(self.model, self.agg_mode)
            if tracker:
                tracker.record_batch_agg()
            loss = self.optimizer.step(closure=closure)
            if tracker:
                tracker.record_batch_opt_step()
        else:
            loss = self.optimizer.step(closure=closure)
            if tracker:
                tracker.record_batch_opt_step()
            self.agg(self.model, self.agg_mode)
            if tracker:
                tracker.record_batch_agg()
        return loss

    def __getattr__(self, item):
        """Any call to unknown function/members will be redirected to the optimizer"""
        if self.optimizer is None:
            raise ValueError("Undefined Optimizer")
        return self.optimizer.__getattribute__(item)


class CentralizedSparsifiedSGD(SparsifiedSGD):
    """Implements centralized sparsified version of stochastic gradient descent.

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): Learning rate
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        sparse_grad_size (int): Size of the sparsified gradients vector (default: 10)
        random_sparse (bool): Whether select random sparsification (default: `False`)
        average_world (bool): Whether to average models on the world_size (default: `True`)

    """

    def __init__(
        self,
        params=None,
        lr=required,
        weight_decay=0,
        sparse_grad_size=10,
        random_sparse=False,
        average_world=True,
    ):
        if not params:
            raise ValueError('"params" not set for optimizer')
        self.average_world = average_world
        self.world_size = dist.get_world_size()
        self.random_sparse = random_sparse
        super(CentralizedSparsifiedSGD, self).__init__(
            params, lr, weight_decay, sparse_grad_size
        )

    def step(self, closure=None):
        """Aggregates the gradients and performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model and returns the loss.
        """

        loss = None

        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            weight_decay = group["weight_decay"]
            lr = group["lr"]

            for p in group["params"]:
                # Sparsify the gradients
                sparse_tensor = self.sparsify_gradients(p, lr)
                # Aggregate the gradients
                gathered_list = [
                    torch.zeros_like(sparse_tensor) for _ in range(self.world_size)
                ]
                dist.all_gather(gathered_list, sparse_tensor)
                p.grad.data = torch.zeros_like(p.grad.data)

                if self.random_sparse:
                    for grad_tensor in gathered_list:
                        for index in range(grad_tensor.size()[1]):
                            p.grad.data[0, int(grad_tensor[0, index])] += grad_tensor[
                                1, index
                            ]
                else:
                    for grad_tensor in gathered_list:
                        tensor_size = grad_tensor.size()[1]
                        begin = int(grad_tensor[0, 0])
                        p.grad.data[
                            0, begin : (begin + tensor_size - 1)
                        ] += grad_tensor[0, 1:]

                if self.average_world:
                    p.grad.data /= self.world_size

                if p.grad is None:
                    continue
                d_p = p.grad.data

                if weight_decay != 0:
                    d_p.add_(weight_decay, p.data)
                p.data.add_(-d_p)

        return loss


class CentralizedSGD(GenericCentralizedOptimizer):
    """Implements centralized stochastic gradient descent (optionally with momentum).
    Averages the reduced parameters over the world size.

    Args:
        world_size (int): Size of the network
        model (:obj:`nn.Module`): Model which contains parameters for SGD
        lr (float): learning rate
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)
        use_cuda (bool): Whether to use cuda tensors for aggregation
        by_layer (bool): Aggregate by layer instead of all layers at once
        agg_grad (bool): Aggregate the gradients before updating weights. If `False`,
            weights will be updated and then reduced across all workers. (default: `True`)
    """

    def __init__(
        self,
        world_size=None,
        model=None,
        lr=required,
        momentum=0,
        dampening=0,
        weight_decay=0,
        nesterov=False,
        use_cuda=False,
        by_layer=False,
        agg_grad=True,
    ):
        super().__init__(
            model=model,
            world_size=world_size,
            use_cuda=use_cuda,
            by_layer=by_layer,
            agg_grad=agg_grad,
        )

        self.optimizer = SGD(
            lr=lr,
            momentum=momentum,
            dampening=dampening,
            weight_decay=weight_decay,
            nesterov=nesterov,
            params=model.parameters(),
        )


class CentralizedAdam(GenericCentralizedOptimizer):
    """Implements centralized Adam algorithm.
    Averages the reduced parameters over the world size

    Args:
        world_size (int): Size of the network
        model (:obj:`nn.Module`): Model which contains parameters for Adam
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper :cite:`adam_convergence`
            (default: False)
        use_cuda (bool): Whether to use cuda tensors for aggregation
        by_layer (bool): Aggregate by layer instead of all layers at once
    """

    def __init__(
        self,
        world_size=None,
        model=None,
        lr=required,
        betas=(0.9, 0.999),
        eps=1e-8,
        weight_decay=0,
        amsgrad=False,
        use_cuda=False,
        by_layer=False,
        agg_grad=True,
    ):
        super(GenericCentralizedOptimizer, self).__init__(
            model=model,
            world_size=world_size,
            use_cuda=use_cuda,
            by_layer=by_layer,
            agg_grad=agg_grad,
        )

        self.optimizer = Adam(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            amsgrad=amsgrad,
            params=model.parameters(),
        )


class PowerSGD(SGD):
    """Implements PowerSGD with error feedback (optionally with momentum).

    Args:
        model (:obj:`nn.Module`): Model which contains parameters for SGD
        lr (float): learning rate
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)
        average_world (bool): Whether to average models on the world_size (default: `True`)
        use_cuda (bool): Whether to use cuda tensors for aggregation
        by_layer (bool): Aggregate by layer instead of all layers at once
        reuse_query (bool): Whether to use warm start to initialize the power iteration
        rank (int): The rank of the gradient approximation
    """

    def __init__(
        self,
        model=None,
        lr=required,
        momentum=0,
        weight_decay=0,
        dampening=0,
        nesterov=False,
        average_world=True,
        use_cuda=False,
        by_layer=False,
        reuse_query=False,
        rank=1,
    ):
        if not model:
            raise ValueError('"model" not set for optimizer')
        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        super(PowerSGD, self).__init__(
            model.parameters(), lr, momentum, dampening, weight_decay, nesterov
        )
        if average_world:
            self.agg_mode = "avg"
        else:
            raise NotImplementedError("Only average model is supported right now.")

        self.model = model
        self.agg = PowerAggregation(
            model=model, use_cuda=use_cuda, reuse_query=reuse_query, rank=rank
        ).agg_grad(by_layer=by_layer)

    def step(self, closure=None, tracker=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
            tracker (:obj:`mlbench_core.utils.Tracker`, optional) The current tracker
        """
        self.agg(self.model, self.agg_mode)
        if tracker:
            tracker.record_batch_agg()
        loss = super(PowerSGD, self).step(closure=closure)
        if tracker:
            tracker.record_batch_opt_step()
        return loss


class CustomCentralizedOptimizer(GenericCentralizedOptimizer):
    """Custom Centralized Optimizer. Can be used with any optimizer passed as argument.
    Adds the gradient clipping option, as well as custom average

    Args:
        model (:obj:`torch.nn.Module`): model
        world_size (int): Distributed world size
        use_cuda (bool): Use cuda tensors for aggregation
        by_layer (bool): Aggregate by layer
        agg_grad (bool): Aggregate the gradients before updating weights. If `False`,
            weights will be updated and then reduced across all workers. (default: `True`)
        grad_clip (float): coefficient for gradient clipping, max L2 norm of the gradients
        average_world (bool): Average the gradients by world size
        average_custom (bool): Divide gradients by given denominator at each step, instead
            of `world_size`
        divide_before (bool): Divide gradients before reduction (default: False)
    """

    def __init__(
        self,
        model,
        world_size,
        optimizer,
        use_cuda=False,
        by_layer=False,
        agg_grad=True,
        grad_clip=float("inf"),
        average_world=False,
        average_custom=False,
        divide_before=False,
    ):
        super(GenericCentralizedOptimizer, self).__init__(
            model=model,
            world_size=world_size,
            use_cuda=use_cuda,
            by_layer=by_layer,
            agg_grad=agg_grad,
            divide_before=divide_before,
        )
        self.grad_clip = grad_clip

        if average_world:
            self.agg_mode = AVG_WORLD
        elif average_custom:
            self.agg_mode = AVG_CUSTOM
        else:
            raise NotImplementedError("Please choose aggregation method")

        self.optimizer = optimizer

    def step(self, closure=None, tracker=None, denom=None):
        """Performs one step of the optimizer.

        Args:
            closure (callable): Optimizer closure argument
            tracker (:obj:`mlbench_core.utils.Tracker`, optional) The current tracker
            denom (Optional[:obj:`torch.Tensor`]): Custom denominator to reduce by
        """

        if self.agg_grad:
            self.agg(self.model, self.agg_mode)
            if tracker:
                tracker.record_batch_agg()

            # Clip norm
            if self.grad_clip != float("inf"):
                clip_grad_norm_(self.model.parameters(), self.grad_clip)

            self.optimizer.step(closure=closure)
            if tracker:
                tracker.record_batch_opt_step()
        else:
            self.optimizer.step(closure=closure)
            if tracker:
                tracker.record_batch_opt_step()
            self.agg(self.model, self.agg_mode)
            if tracker:
                tracker.record_batch_agg()

        return True



import os
from copy import deepcopy

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm

from mlbench_core.dataset.util.tools import maybe_download_and_extract_tar_gz

from .wmt16 import (
    WMT16Tokenizer,
    filter_data,
    filter_raw_data,
    get_data_dtype,
    process_data,
    process_raw_data,
    wmt16_config,
)


class WMT16Dataset(Dataset):
    """Dataset for WMT16 en to de translation

    Args:
        root (str): Root folder where to download files
        lang (tuple): Language translation pair
        math_precision (str): One of `fp16` or `fp32`. The precision used during training
        download (bool): Download the dataset from source
        train (bool): Load train set
        validation (bool): Load validation set
        lazy (bool): Load the dataset in a lazy format
        min_len (int): Minimum sentence length
        max_len (int | None): Maximum sentence length
        max_size (int | None): Maximum dataset size
    """

    urls = [
        (
            "https://storage.googleapis.com/mlbench-datasets/translation/wmt16_en_de.tar.gz",
            "wmt16_en_de.tar.gz",
        )
    ]

    def __init__(
        self,
        root,
        lang=("en", "de"),
        math_precision=None,
        download=True,
        train=False,
        validation=False,
        lazy=False,
        preprocessed=False,
        sort=False,
        min_len=0,
        max_len=None,
        max_size=None,
    ):
        super(WMT16Dataset, self).__init__()
        if download:
            url, file_name = self.urls[0]
            maybe_download_and_extract_tar_gz(root, file_name, url)

        if train:
            self.path = os.path.join(root, wmt16_config.TRAIN_FNAME)
        elif validation:
            self.path = os.path.join(root, wmt16_config.VAL_FNAME)
        else:
            raise NotImplementedError()

        self.min_len, self.max_len = min_len, max_len
        self.lazy = lazy
        self.preprocessed = preprocessed
        self.tokenizer = WMT16Tokenizer(root, math_precision=math_precision)

        # Data is not tokenized already
        if not preprocessed:
            src_path, trg_path = tuple(
                os.path.expanduser(self.path + "." + x) for x in lang
            )
            if lazy:
                raw_src = process_raw_data(src_path, max_size)
                raw_trg = process_raw_data(trg_path, max_size)
                assert len(raw_src) == len(
                    raw_trg
                ), "Source and target have different lengths"

                raw_src, src_lens, raw_trg, trg_lens = filter_raw_data(
                    raw_src, raw_trg, min_len - 2, max_len - 2
                )

                assert len(raw_src) == len(
                    raw_trg
                ), "Source and target have different lengths"
                # Adding 2 because EOS and BOS are added later during tokenization
                src_lengths = [i + 2 for i in src_lens]
                trg_lengths = [i + 2 for i in trg_lens]
                self.src = raw_src
                self.trg = raw_trg
            else:
                src = process_data(src_path, self.tokenizer, max_size)
                trg = process_data(trg_path, self.tokenizer, max_size)
                assert len(src) == len(trg), "Source and target have different lengths"

                src, trg = filter_data(src, trg, min_len, max_len)

                assert len(src) == len(trg), "Source and target have different lengths"

                src_lengths = [len(s) for s in src]
                trg_lengths = [len(t) for t in trg]
                self.src, self.trg = src, trg
        else:
            self.path = "{}.bin".format(self.path)
            self.file = None
            (
                file_max_len,
                file_min_len,
                file_vocab_size,
                src_lengths,
                trg_lengths,
            ) = self._extract_preprocessed_data()

            assert file_max_len == self.max_len
            assert file_min_len == self.min_len
            assert file_vocab_size == self.vocab_size

            self.dtype, _ = get_data_dtype(self.vocab_size)
            itemsize = np.iinfo(self.dtype).dtype.itemsize
            self.item_stride = itemsize * self.max_len * 2

        self.src_lengths = torch.tensor(src_lengths)
        self.trg_lengths = torch.tensor(trg_lengths)
        # self.lengths = self.src_lengths + self.trg_lengths

        self.sorted = False
        if sort:
            self._sort_by_src_length()
            self.sorted = True

    def _extract_preprocessed_data(self):
        if not (os.path.exists(self.path) and os.path.isfile(self.path)):
            raise ValueError("File {} not found".format(self.path))
        with open(self.path, "rb") as f:
            length = int(np.fromfile(f, np.int64, 1))
            file_vocab_size = int(np.fromfile(f, np.int64, 1))
            file_min_len = int(np.fromfile(f, np.int64, 1))
            file_max_len = int(np.fromfile(f, np.int64, 1))

            src_lengths = np.fromfile(f, np.int64, length)
            trg_lengths = np.fromfile(f, np.int64, length)
            self.offset = int(np.fromfile(f, np.int64, 1))
        return file_max_len, file_min_len, file_vocab_size, src_lengths, trg_lengths

    def _sort_by_src_length(self):
        self.src_lengths, indices = self.src_lengths.sort(descending=True)
        self.src = [self.src[i] for i in indices]
        self.trg = [self.trg[i] for i in indices]
        self.trg_lengths = [self.trg_lengths[i] for i in indices]

    def prepare(self):
        if self.preprocessed:
            self.file = open(self.path, "rb")

    def __getitem__(self, idx):
        if self.preprocessed:
            offset = self.offset + self.item_stride * idx
            self.file.seek(offset, os.SEEK_SET)
            data = np.fromfile(self.file, self.dtype, self.max_len * 2)
            data = data.astype(np.int64)
            src_len = self.src_lengths[idx]
            tgt_len = self.trg_lengths[idx]
            src = torch.tensor(data[0:src_len])
            tgt = torch.tensor(data[self.max_len : self.max_len + tgt_len])
            return src, tgt
        else:
            if self.lazy:
                src = torch.tensor(self.tokenizer.segment(self.src[idx]))
                trg = torch.tensor(self.tokenizer.segment(self.trg[idx]))
            else:
                src, trg = self.src[idx], self.trg[idx]
            return src, trg

    def write_as_preprocessed(
        self, collate_fn, min_len=0, max_len=50, num_workers=2, batch_size=1024
    ):
        loader = DataLoader(
            self,
            batch_size=batch_size,
            collate_fn=collate_fn,
            num_workers=num_workers,
            drop_last=False,
        )

        srcs = []
        tgts = []
        src_lengths = []
        tgt_lengths = []
        dtype, torch_dtype = get_data_dtype(self.vocab_size)

        for (src, src_len), (tgt, tgt_len) in tqdm(loader):
            src_lengths.append(deepcopy(src_len.to(torch_dtype)))
            tgt_lengths.append(deepcopy(tgt_len.to(torch_dtype)))
            srcs.append(deepcopy(src))
            tgts.append(deepcopy(tgt))
            del src
            del src_len
            del tgt
            del tgt_len

        del loader
        import gc

        gc.collect()

        srcs = torch.cat(srcs)
        tgts = torch.cat(tgts)
        src_lengths = torch.cat(src_lengths)
        tgt_lengths = torch.cat(tgt_lengths)

        assert len(srcs) == len(tgts) == len(src_lengths) == len(tgt_lengths)
        length = len(srcs)

        offset = 0
        dest_fname = "{}.bin".format(self.path)
        with open(dest_fname, "wb") as f:
            offset += f.write((np.array(length, dtype=np.int64)))
            offset += f.write((np.array(self.vocab_size, dtype=np.int64)))
            offset += f.write((np.array(min_len, dtype=np.int64)))
            offset += f.write((np.array(max_len, dtype=np.int64)))
            offset += f.write((np.array(src_lengths, dtype=np.int64)))
            offset += f.write((np.array(tgt_lengths, dtype=np.int64)))

            offset += np.iinfo(np.int64).dtype.itemsize
            f.write((np.array(offset, dtype=np.int64)))

            del src_lengths
            del tgt_lengths
            gc.collect()
            data = torch.cat((srcs, tgts), dim=1).numpy()
            f.write((np.array(data, dtype=dtype)))

    @property
    def vocab_size(self):
        return self.tokenizer.vocab_size

    def __len__(self):
        return len(self.src_lengths)



from .log_metrics import LogMetrics
from .tracker import AverageMeter, Tracker

try:
    import torch

    from . import pytorch
except ImportError:
    pass


try:
    import tensorflow

    from . import tensorflow
except ImportError:
    pass



"""Tests for `mlbench_core.utils` package."""

import datetime

from freezegun import freeze_time

from mlbench_core.evaluation.goals import task1_time_to_accuracy_light_goal
from mlbench_core.evaluation.pytorch.metrics import TopKAccuracy
from mlbench_core.utils import LogMetrics, Tracker


def test_tracker():
    tracker = Tracker([TopKAccuracy(5)], 1, 0)

    assert tracker is not None


def test_tracker_goal(mocker):
    patched = mocker.patch("mlbench_core.utils.tracker.LogMetrics")

    metric = TopKAccuracy(1)
    tracker = Tracker([metric], 1, 0, task1_time_to_accuracy_light_goal())

    tracker.start()

    assert tracker.start_time is not None

    tracker.train()

    tracker.record_stat("global_Prec@1", 69, log_to_api=True)
    tracker.batch_end()

    assert not tracker.goal_reached

    tracker.record_stat("global_Prec@1", 70, log_to_api=True)
    tracker.batch_end()

    assert not tracker.goal_reached

    tracker.validation()

    tracker.record_stat("global_Prec@1", 69, log_to_api=True)
    tracker.batch_end()

    assert not tracker.goal_reached

    tracker.record_stat("global_Prec@1", 70, log_to_api=True)

    assert tracker.goal_reached


def _do_batch(tracker, frozen):
    tracker.batch_start()
    frozen.tick(delta=datetime.timedelta(seconds=0.5))
    tracker.record_batch_load()
    frozen.tick(delta=datetime.timedelta(seconds=0.5))
    tracker.record_batch_init()
    frozen.tick(delta=datetime.timedelta(seconds=0.5))
    tracker.record_batch_fwd_pass()
    frozen.tick(delta=datetime.timedelta(seconds=0.5))
    tracker.record_batch_comp_loss()
    frozen.tick(delta=datetime.timedelta(seconds=0.5))
    tracker.record_batch_backprop()
    frozen.tick(delta=datetime.timedelta(seconds=0.5))
    tracker.record_batch_agg()
    frozen.tick(delta=datetime.timedelta(seconds=0.5))
    tracker.record_batch_opt_step()
    frozen.tick(delta=datetime.timedelta(seconds=0.5))
    tracker.record_batch_comp_metrics()
    frozen.tick(delta=datetime.timedelta(seconds=0.5))
    tracker.batch_end()


def test_tracker_goal_times(mocker):
    patched = mocker.patch("mlbench_core.utils.tracker.LogMetrics")

    metric = TopKAccuracy(1)
    tracker = Tracker([metric], 1, 0, task1_time_to_accuracy_light_goal())

    tracker.start()

    assert tracker.start_time is not None

    tracker.train()

    with freeze_time(datetime.datetime.now()) as frozen:
        _do_batch(tracker, frozen)

        assert abs(tracker.get_total_preprocess_time() - 0.5) < 0.01
        assert abs(tracker.get_total_communication_time() - 0.5) < 0.01
        assert abs(tracker.get_total_compute_time() - 2.0) < 0.01
        assert abs(tracker.get_total_metrics_time() - 0.5) < 0.01

        _do_batch(tracker, frozen)

        assert abs(tracker.get_total_preprocess_time() - 1.0) < 0.01
        assert abs(tracker.get_total_communication_time() - 1.0) < 0.01
        assert abs(tracker.get_total_compute_time() - 4.0) < 0.01
        assert abs(tracker.get_total_metrics_time() - 1.0) < 0.01

        tracker.validation()
        tracker.record_stat("global_Prec@1", 70, log_to_api=True)

        assert tracker.goal_reached
        assert any(filter(lambda c: c[1][3] == "TaskResult", patched.method_calls))


def test_LogMetrics(mocker):
    mocker.patch("mlbench_core.api.ApiClient")

    LogMetrics.log("1", 1, 1, "loss", 123)

    mocker.patch.dict("os.environ", {"MLBENCH_IN_DOCKER": "True"})

    LogMetrics.log("1", 1, 1, "loss", 123)



import os
import subprocess
import sys
from time import sleep
from urllib import request

import click
import google.auth
import yaml
from google.api_core.exceptions import AlreadyExists, NotFound
from google.auth.exceptions import DefaultCredentialsError
from google.cloud import container_v1
from googleapiclient import discovery, http
from kubernetes import client as kube_client

GCLOUD_NVIDIA_DAEMONSET = (
    "https://raw.githubusercontent.com/"
    "GoogleCloudPlatform/container-engine-accelerators/"
    "stable/nvidia-driver-installer/cos/"
    "daemonset-preloaded.yaml"
)


def _create_kube_config_gcloud_entry(cluster_name, cluster_zone, project):
    """Uses GCloud CLI to create an entry for Kubectl.

    This is needed as we install the charts using kubectl, and it needs the correct config

    Args:
        cluster_name (str): Name of cluster
        cluster_zone (str): Zone of cluster
        project (str): Current used project

    Returns:
        (str): Kube context for the cluster
    """
    p = subprocess.Popen(
        [
            "gcloud",
            "container",
            "clusters",
            "get-credentials",
            cluster_name,
            "--zone",
            cluster_zone,
            "--project",
            project,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    output, error = p.communicate()

    if p.returncode != 0:
        raise click.UsageError(
            "Failed to add kube config entry:\n {}".format(error.decode())
        )

    context_name = "gke_{}_{}_{}".format(project, cluster_zone, cluster_name)
    return context_name


def setup_gcloud_kube_client(cluster_endpoint, cluster_name, cluster_zone, project):
    """Sets up the kube configuration on the given cluster

    Args:
        cluster_endpoint (str): Endpoint of cluster
        cluster_name (str): Cluster name
        cluster_zone (str): Cluster zone
        project (str): Cluster project

    Returns:
        (str): Kube context for this cluster
    """
    try:
        credentials, _ = google.auth.default()
    except DefaultCredentialsError:
        raise click.UsageError(
            "Couldn't find gcloud credentials. Install the gcloud"
            " sdk ( https://cloud.google.com/sdk/docs/quickstart-linux ) and "
            "run 'gcloud auth application-default login' to login and create "
            "your credentials."
        )

    # Setup kube client
    auth_req = google.auth.transport.requests.Request()
    credentials.refresh(auth_req)
    configuration = kube_client.Configuration()
    configuration.host = f"https://{cluster_endpoint}:443"
    configuration.verify_ssl = False
    configuration.api_key = {"authorization": "Bearer " + credentials.token}
    kube_client.Configuration.set_default(configuration)

    kube_context = _create_kube_config_gcloud_entry(cluster_name, cluster_zone, project)

    return kube_context


def gcloud_create_cluster(
    name,
    name_path,
    num_workers,
    num_gpus,
    gpu_type,
    machine_type,
    disk_size,
    preemptible,
    kubernetes_version,
    project,
):
    """Creates a GCloud cluster based on the given arguments

    Args:
        name (str): Cluster name
        name_path (str): Cluster name path (` "projects/{}/locations/{}/clusters/"`)
        num_workers (int): Total number of nodes (including master node)
        num_gpus (int): Total number of gpus per node
        gpu_type (str): GPU type
        machine_type (str): Machine type
        disk_size (int): Disk size in GB
        preemptible (bool): Use preemptible machines
        kubernetes_version (str): Kubernetes version
        project (str): GCLoud project

    Returns:
        (:obj:`google.container_v1.ClusterManagementClient`, str, :obj:): The client for cluster communication,
            firewall name, and cluster firewalls
    """
    assert num_workers >= 2, "Number of workers should be at least 2"

    # create cluster
    gclient = container_v1.ClusterManagerClient()

    extraargs = {}

    if num_gpus > 0:
        extraargs["accelerators"] = [
            container_v1.types.AcceleratorConfig(
                accelerator_count=num_gpus, accelerator_type=gpu_type
            )
        ]

    # delete existing firewall, if any
    firewalls = discovery.build("compute", "v1", cache_discovery=False).firewalls()
    existing_firewalls = firewalls.list(project=project).execute()
    fw_name = "{}-firewall".format(name)

    if any(f["name"] == fw_name for f in existing_firewalls["items"]):
        response = {}
        while not hasattr(response, "status"):
            try:
                response = firewalls.delete(project=project, firewall=fw_name).execute()
            except http.HttpError as e:
                if e.resp.status == 404:
                    response = {}
                    break
                click.echo("Wait for firewall to be available for deletion")
                sleep(5)
                response = {}
        while hasattr(response, "status") and response.status < response.DONE:
            response = gclient.get_operation(None, None, None, name=response.selfLink)
            sleep(1)

    # create cluster
    cluster = container_v1.types.Cluster(
        name=name,
        initial_node_count=num_workers,
        node_config=container_v1.types.NodeConfig(
            machine_type=machine_type,
            disk_size_gb=disk_size,
            preemptible=preemptible,
            oauth_scopes=[
                "https://www.googleapis.com/auth/devstorage.full_control",
            ],
            **extraargs,
        ),
        addons_config=container_v1.types.AddonsConfig(
            http_load_balancing=container_v1.types.HttpLoadBalancing(
                disabled=True,
            ),
            horizontal_pod_autoscaling=container_v1.types.HorizontalPodAutoscaling(
                disabled=True,
            ),
            kubernetes_dashboard=container_v1.types.KubernetesDashboard(
                disabled=True,
            ),
            network_policy_config=container_v1.types.NetworkPolicyConfig(
                disabled=False,
            ),
        ),
        logging_service=None,
        monitoring_service=None,
        initial_cluster_version=kubernetes_version,
    )
    try:
        response = gclient.create_cluster(cluster, parent=name_path)
    except AlreadyExists as e:
        click.echo("Exception from Google: " + str(e))
        click.echo(
            "A cluster with this name already exists in the specified project and zone"
        )
        click.echo(
            "Try running 'gcloud container clusters list' to list all active clusters"
        )
        sys.exit(1)

    # wait for cluster to load
    while response.status < response.DONE:
        response = gclient.get_operation(
            None, None, None, name=os.path.join(name_path, response.name)
        )
        sleep(1)

    if response.status != response.DONE:
        raise ValueError("Cluster creation failed!")

    return gclient, fw_name, firewalls


def deploy_nvidia_daemonset():
    """ Deploys the NVIDIA daemon set to the cluster"""
    with request.urlopen(GCLOUD_NVIDIA_DAEMONSET) as r:
        dep = yaml.safe_load(r)
        dep["spec"]["selector"] = {
            "matchLabels": dep["spec"]["template"]["metadata"]["labels"]
        }
        dep = kube_client.ApiClient()._ApiClient__deserialize(dep, "V1DaemonSet")
        k8s_client = kube_client.AppsV1Api()
        k8s_client.create_namespaced_daemon_set("kube-system", body=dep)


def gcloud_delete_cluster(name, zone, project):
    """Deletes the cluster

    Args:
        name (str): Cluster name
        zone (str): Cluster zone
        project (str): Cluster project
    """
    try:
        credentials, default_project = google.auth.default()
    except DefaultCredentialsError:
        raise click.UsageError(
            "Couldn't find gcloud credentials. Install the gcloud"
            " sdk ( https://cloud.google.com/sdk/docs/quickstart-linux ) and "
            "run 'gcloud auth application-default login' to login and create "
            "your credentials."
        )

    if not project:
        project = default_project

    # delete cluster
    gclient = container_v1.ClusterManagerClient()

    name_path = "projects/{}/locations/{}/clusters/".format(project, zone)
    cluster_path = os.path.join(name_path, name)

    try:
        response = gclient.delete_cluster(None, None, None, name=cluster_path)
    except NotFound as e:
        click.echo("Exception from Google: " + str(e))
        click.echo("Double-check your project, zone and cluster name")
        click.echo(
            "Try running 'gcloud container clusters list' to list all active clusters"
        )
        sys.exit(1)

    # wait for operation to complete
    while response.status < response.DONE:
        response = gclient.get_operation(
            None, None, None, name=os.path.join(name_path, response.name)
        )
        sleep(1)

    if response.status != response.DONE:
        raise ValueError("Cluster deletion failed!")



import logging
import os
import struct

import numpy as np
import torch
from torch.utils.data import Dataset

from mlbench_core.dataset.util.tools import maybe_download_and_extract_tar_gz

from .wmt17 import Dictionary, collate_batch

logger = logging.getLogger("mlbench")

dtypes = {
    1: np.uint8,
    2: np.int8,
    3: np.int16,
    4: np.int32,
    5: np.int64,
    6: np.float,
    7: np.double,
}


def _read_longs(f, n):
    a = np.empty(n, dtype=np.int64)
    f.readinto(a)
    return a


def _index_file_path(prefix_path):
    return prefix_path + ".idx"


def _data_file_path(prefix_path):
    return prefix_path + ".bin"


class WMT17Dataset(Dataset):
    """Dataset for WMT17 EN to DE translation, for transformer model.

    The dataset needs to be pre-processed before training by using the script
    `mlbench_core/dataset/nlp/pytorch/wmt17/preprocess/preprocess.py`.

    This class uses the `.bin` and `.idx` files, outputted by the pre-processing script

    Args:
        root (str): Root folder where to download files
        lang (tuple): Language pair
        download (bool): Download dataset from gcloud S3
        train (bool): Load train set
        validation (bool): Load validation set
        test (bool): Load test set
        left_pad (tuple[bool]): left- or right-padding (true: left, false: right) for (source, target)
        max_positions (tuple[int]): Maximum number of tokens in (source, target)
        seq_len_multiple (int): Pad sources to multiples of this
        shuffle (bool): Shuffle dataset
    """

    urls = [
        (
            "https://storage.googleapis.com/mlbench-datasets/translation/wmt17_en_de.tar.gz",
            "wmt17_en_de.tar.gz",
        )
    ]
    name = "wmt17"
    dirname = ""

    prefixes = {"train": "train", "validation": "dev", "test": "test"}

    def __init__(
        self,
        root,
        lang=("en", "de"),
        download=False,
        train=False,
        validation=False,
        test=False,
        left_pad=(True, False),
        max_positions=(256, 256),
        seq_len_multiple=1,
        shuffle=True,
    ):
        src_lang, trg_lang = lang
        self.left_pad_source, self.left_pad_target = left_pad
        self.max_source_positions, self.max_target_positions = max_positions

        self.seq_len_multiple = seq_len_multiple
        self.shuffle = shuffle
        if download:
            url, file_name = self.urls[0]
            maybe_download_and_extract_tar_gz(root, file_name, url)

        self.src_dict = Dictionary.load(
            os.path.join(root, "dict.{}.txt".format(src_lang))
        )
        self.trg_dict = Dictionary.load(
            os.path.join(root, "dict.{}.txt".format(trg_lang))
        )

        if train:
            self.prefix = self.prefixes["train"]
        elif validation:
            self.prefix = self.prefixes["validation"]
        elif test:
            self.prefix = self.prefixes["test"]
        else:
            raise NotImplementedError()

        assert self.src_dict.pad() == self.trg_dict.pad()
        assert self.src_dict.eos() == self.trg_dict.eos()

        self.src_path = os.path.join(
            root, "{}.{}-{}.{}".format(self.prefix, src_lang, trg_lang, src_lang)
        )
        self.trg_path = os.path.join(
            root, "{}.{}-{}.{}".format(self.prefix, src_lang, trg_lang, trg_lang)
        )

        assert self._exists()

        self.src_data = IndexedDataset(self.src_path)
        self.trg_data = IndexedDataset(self.trg_path)

        self.src_sizes = np.array(self.src_data.sizes)
        self.trg_sizes = np.array(self.trg_data.sizes)

        print(
            "| Sentences are being padded to multiples of: {}".format(
                self.seq_len_multiple
            )
        )

    def __len__(self):
        return len(self.src_data)

    def _exists(self):
        """Checks if all the files needed exist

        Returns:
            (bool): True of all the files exist
        """
        return (
            os.path.exists(_index_file_path(self.src_path))
            and os.path.exists(_data_file_path(self.src_path))
            and os.path.exists(_index_file_path(self.trg_path))
            and os.path.join(_data_file_path(self.trg_path))
        )

    def __getitem__(self, index):
        return {
            "id": index,
            "source": self.src_data[index],
            "target": self.trg_data[index],
        }

    def collater(self, samples):
        """Merge a list of samples to form a mini-batch."""
        return collate_batch(
            samples,
            pad_idx=self.src_dict.pad(),
            eos_idx=self.src_dict.eos(),
            left_pad_source=self.left_pad_source,
            left_pad_target=self.left_pad_target,
            bsz_mult=8,
            seq_len_multiple=self.seq_len_multiple,
        )

    def ordered_indices(self, partition_indices=None, seed=None):
        """Returns a list of indices, ordered by source and target lengths

        If `partition_indices` is not `None`, only those indices will be returned,
        but sorted.

        The indices will be shuffled before sorting, using the given seed

        Args:
            partition_indices (Optional[list[int]]): The list of indices to sort
                Default: `None`
            seed (Optional[int]): Seed to use for shuffling

        Returns:
            (`obj`:np.array): Array of sorted indices
        """
        if self.shuffle:
            indices = np.random.RandomState(seed).permutation(
                len(self) if partition_indices is None else partition_indices
            )
        else:
            indices = (
                np.arange(len(self)) if partition_indices is None else partition_indices
            )

        if self.trg_sizes is not None:
            indices = indices[np.argsort(self.trg_sizes[indices], kind="mergesort")]

        return indices[np.argsort(self.src_sizes[indices], kind="mergesort")]


class IndexedDataset(Dataset):
    """Loader for TorchNet IndexedDataset"""

    def __init__(self, path):
        super().__init__()
        # Read Index file
        with open(_index_file_path(path), "rb") as f:
            magic = f.read(8)
            assert magic == b"TNTIDX\x00\x00"
            version = f.read(8)
            assert struct.unpack("<Q", version) == (1,)
            code, self.element_size = struct.unpack("<QQ", f.read(16))
            self.dtype = dtypes[code]
            self.size, self.s = struct.unpack("<QQ", f.read(16))
            self.dim_offsets = _read_longs(f, self.size + 1)
            self.data_offsets = _read_longs(f, self.size + 1)
            self.sizes = _read_longs(f, self.s)
        with open(_data_file_path(path), "rb") as f:
            self.buffer = np.empty(self.data_offsets[-1], dtype=self.dtype)
            f.readinto(self.buffer)

    def check_index(self, i):
        if i < 0 or i >= self.size:
            raise IndexError("index out of range")

    def __getitem__(self, i):
        self.check_index(i)
        tensor_size = self.sizes[self.dim_offsets[i] : self.dim_offsets[i + 1]]
        a = np.empty(tensor_size, dtype=self.dtype)
        np.copyto(a, self.buffer[self.data_offsets[i] : self.data_offsets[i + 1]])
        return torch.from_numpy(a).long()

    def __len__(self):
        return self.size



import os
import subprocess
import tempfile
from time import sleep

import click
import docker

KIND_NODES_IMAGES = {
    "1.15": "kindest/node:v1.15.12@sha256:d9b939055c1e852fe3d86955ee24976cab46cba518abcb8b13ba70917e6547a6",
    "1.16": "kindest/node:v1.16.15@sha256:a89c771f7de234e6547d43695c7ab047809ffc71a0c3b65aa54eda051c45ed20",
    "1.17": "kindest/node:v1.17.11@sha256:5240a7a2c34bf241afb54ac05669f8a46661912eab05705d660971eeb12f6555",
    "1.18": "kindest/node:v1.18.8@sha256:f4bcc97a0ad6e7abaf3f643d890add7efe6ee4ab90baeb374b4f41a4c95567eb",
    "1.19": "kindest/node:v1.19.1@sha256:98cf5288864662e37115e362b23e4369c8c4a408f99cbc06e58ac30ddc721600",
}
KIND_CONFIG = """
kind: Cluster
apiVersion: kind.x-k8s.io/v1alpha4
containerdConfigPatches:
- |-
  [plugins."io.containerd.grpc.v1.cri".registry.mirrors."localhost:{reg_port}"]
    endpoint = ["http://{reg_name}:{reg_port}"]
nodes:
- role: control-plane
  image: {kind_node_image}
{workers_config}
"""


def create_local_registry(registry_name, registry_port, host_port):
    """Creates a local docker registry

    Args:
        registry_name (str): Name of registry
        registry_port (str): Port registry
        host_port (str): Host port to bind registry to
    """
    docker_client = docker.from_env()

    # check if local registry exists
    existing_containers = [
        container.name for container in docker_client.containers.list()
    ]
    registry_exists = registry_name in existing_containers

    if not registry_exists:
        # create local registry
        click.echo("Creating registry {}".format(registry_name))
        docker_client.containers.run(
            image="registry:2",
            name=registry_name,
            restart_policy={"Name": "always"},
            ports={registry_port: host_port},
            detach=True,
        )
        while docker_client.containers.get(registry_name).status != "running":
            if docker_client.containers.get(registry_name).status == "exited":
                raise click.UsageError("Failed to create local registry")
            sleep(1)


def kind_create_cluster(
    name, num_workers, registry_name, registry_port, kubernetes_version
):
    """Creates a local KIND cluster with the given name and connects it to the running local registry

    Args:
        name (str): Cluster name
        num_workers (int): Total number of nodes (including master node)
        registry_name (str): Local registry name
        registry_port (str): Local registry port
        kubernetes_version (str): Kubernetes version to use (available :1.15, 1.16, 1.17, 1.18, 1.19)
    """
    # create cluster
    with tempfile.TemporaryDirectory() as temp_directory:
        kind_config_file_location = os.path.join(temp_directory, "kind_config.yml")

        with open(kind_config_file_location, "w") as f:
            workers_config = "\n".join(
                ["- role: worker\n  image: {kind_node_image}"] * (num_workers - 1)
            )
            node_image = KIND_NODES_IMAGES[kubernetes_version]
            f.write(
                KIND_CONFIG.format(
                    reg_port=registry_port,
                    reg_name=registry_name,
                    workers_config=workers_config.format(kind_node_image=node_image),
                    kind_node_image=node_image,
                )
            )

        click.echo("Creating cluster {}".format(name))

        p = subprocess.Popen(
            [
                "kind",
                "create",
                "cluster",
                "--name",
                name,
                "--config",
                kind_config_file_location,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        output, error = p.communicate()
        if p.returncode != 0:
            raise click.UsageError(
                "Failed to create cluster with the following error:\n {}".format(
                    error.decode()
                )
            )

    _connect_to_local_registry(registry_name)


def kind_delete_cluster(name, registry_name):
    """Deletes the cluster using the kind command `kind delete cluster`
    Also disconnects from the local registry

    Args:
        name (str): Cluster name
        registry_name (str): Registry name
    """
    # Delete cluster using kind
    p = subprocess.Popen(
        ["kind", "delete", "cluster", "--name", name],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    _, error = p.communicate()
    if p.returncode != 0:
        raise click.UsageError(
            "Failed to delete cluster with the following error:\n {}".format(
                error.decode()
            )
        )

    _disconnect_from_local_registry(registry_name)


def _connect_to_local_registry(registry_name):
    """Connects the kind network in docker to the given registry name

    Args:
        registry_name (str): Local registry name
    """
    docker_client = docker.from_env()
    # Connect local registry to kind cluster
    kind_network = [x for x in docker_client.networks.list() if x.name == "kind"]
    if len(kind_network) == 0:
        raise ValueError("Kind network not found")
    kind_network = kind_network[0]
    kind_network.connect(registry_name)


def _disconnect_from_local_registry(registry_name):
    """Disconnects the kind network in docker to the given registry name

    Args:
        registry_name (str): Local registry name
    """
    # Disconnect from local registry
    docker_client = docker.from_env()

    kind_network = [x for x in docker_client.networks.list() if x.name == "kind"]
    if len(kind_network) == 0:
        raise ValueError("Kind network not found")
    kind_network = kind_network[0]

    kind_network.disconnect(registry_name)



"""Tests for `mlbench_core.utils.pytorch.utils` package."""
import torch

from mlbench_core.utils.pytorch.utils import orthogonalize, pack_tensors, unpack_tensors


def test_orthogonalize():
    m = torch.rand(2, 2)
    identity = torch.eye(2)

    orthogonalize(m)

    # check if m'*m = I
    assert torch.allclose(torch.matmul(m.t(), m), identity, atol=1e-04)


def test_pack_tensors():
    tensors = [torch.rand(2, 2), torch.rand(2, 2)]

    flattened = [y for x in tensors for y in x.view(-1)]

    vec, indices, sizes = pack_tensors(tensors)

    assert vec.tolist() == flattened
    assert indices == [0, 4, 8]
    assert sizes == [(2, 2), (2, 2)]


def test_unpack_tensors():
    tensors = [torch.rand(2, 2), torch.rand(2, 2)]
    vec, indices, sizes = pack_tensors(tensors)

    unpacked = unpack_tensors(vec, indices, sizes)

    assert all((x == y).all() for x, y in zip(tensors, unpacked))



# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree. An additional grant of patent rights
# can be found in the PATENTS file in the same directory.

import math
import re
import sys
import unicodedata

import six
import torch
from torch.nn import functional as F

from mlbench_core.models.pytorch.transformer.decoder import TransformerDecoder


class UnicodeRegex(object):
    """Ad-hoc hack to recognize all punctuation and symbols."""

    def __init__(self):
        punctuation = self.property_chars("P")
        self.nondigit_punct_re = re.compile(r"([^\d])([" + punctuation + r"])")
        self.punct_nondigit_re = re.compile(r"([" + punctuation + r"])([^\d])")
        self.symbol_re = re.compile("([" + self.property_chars("S") + "])")

    def property_chars(self, prefix):
        return "".join(
            six.unichr(x)
            for x in range(sys.maxunicode)
            if unicodedata.category(six.unichr(x)).startswith(prefix)
        )


uregex = UnicodeRegex()


def strip_pad(tensor, pad):
    return tensor[tensor.ne(pad)]


def post_process_prediction(hypo_tokens, alignment, align_dict, tgt_dict, remove_bpe):
    hypo_str = tgt_dict.string(hypo_tokens, remove_bpe)

    assert not align_dict

    return hypo_tokens, hypo_str, alignment


def detokenize_subtokenized_sentence(subtokenized_sentence):
    l1 = " ".join("".join(subtokenized_sentence.strip().split()).split("_"))
    l1 = l1.replace(" ,", ",")
    l1 = l1.replace(" .", ".")
    l1 = l1.replace(" !", "!")
    l1 = l1.replace(" ?", "?")
    l1 = l1.replace(" ' ", "'")
    l1 = l1.replace(" - ", "-")
    l1 = l1.strip()
    return l1


def bleu_tokenize(string):
    r"""Tokenize a string following the official BLEU implementation.
    See https://github.com/moses-smt/mosesdecoder/'
    'blob/master/scripts/generic/mteval-v14.pl#L954-L983
    In our case, the input string is expected to be just one line
    and no HTML entities de-escaping is needed.
    So we just tokenize on punctuation and symbols,
    except when a punctuation is preceded and followed by a digit
    (e.g. a comma/dot as a thousand/decimal separator).
    Note that a numer (e.g. a year) followed by a dot at the end of sentence
    is NOT tokenized,
    i.e. the dot stays with the number because `s/(\p{P})(\P{N})/ $1 $2/g`
    does not match this case (unless we add a space after each sentence).
    However, this error is already in the original mteval-v14.pl
    and we want to be consistent with it.
    Args:
    string: the input string
    Returns:
    a list of tokens
    """
    string = uregex.nondigit_punct_re.sub(r"\1 \2 ", string)
    string = uregex.punct_nondigit_re.sub(r" \1 \2", string)
    string = uregex.symbol_re.sub(r" \1 ", string)
    return string


class UnicodeRegex(object):
    """Ad-hoc hack to recognize all punctuation and symbols."""

    def __init__(self):
        punctuation = self.property_chars("P")
        self.nondigit_punct_re = re.compile(r"([^\d])([" + punctuation + r"])")
        self.punct_nondigit_re = re.compile(r"([" + punctuation + r"])([^\d])")
        self.symbol_re = re.compile("([" + self.property_chars("S") + "])")

    def property_chars(self, prefix):
        return "".join(
            six.unichr(x)
            for x in range(sys.maxunicode)
            if unicodedata.category(six.unichr(x)).startswith(prefix)
        )


class SequenceGenerator(object):
    """Generates translations of a given source sentence.

    Args:
        model (:obj:`torch.nn.Module`): The model to predict on. Should be instance of `TransformerModel`
        src_dict (:obj:`mlbench_core.dataset.nlp.pytorch.wmt17.Dictionary`): Source dictionary
        trg_dict (:obj:`mlbench_core.dataset.nlp.pytorch.wmt17.Dictionary`): Target dictionary
        beam_size (int): Size of the beam. Default 1
        minlen (int): Minimum generation length. Default 1
        maxlen (int): Maximum generation length. If `None`, takes value of model.max_decoder_positions().
            Default `None`
        stop_early (bool): Stop generation immediately after we finalize beam_size
            hypotheses, even though longer hypotheses might have better
            normalized scores. Default `True`
        normalize_scores (bool): Normalize scores by the length of the output. Default `True`
        len_penalty (float): length penalty: <1.0 favors shorter, >1.0 favors longer sentences.
            Default 1
        retain_dropout (bool): Keep dropout layers. Default `False`
        sampling (bool): sample hypotheses instead of using beam search. Default `False`
        sampling_topk (int): sample from top K likely next words instead of all words. Default -1
        sampling_temperature (int): temperature for random sampling. Default 1
    """

    def __init__(
        self,
        model,
        src_dict,
        trg_dict,
        beam_size=1,
        minlen=1,
        maxlen=None,
        stop_early=True,
        normalize_scores=True,
        len_penalty=1,
        retain_dropout=False,
        sampling=False,
        sampling_topk=-1,
        sampling_temperature=1,
    ):
        self.model = model
        self.pad = trg_dict.pad()
        self.eos = trg_dict.eos()
        self.vocab_size = len(trg_dict)
        self.src_dict = src_dict
        self.trg_dict = trg_dict
        self.beam_size = beam_size
        self.minlen = minlen
        max_decoder_len = self.model.max_decoder_positions()
        max_decoder_len -= 1  # we define maxlen not including the EOS marker
        self.maxlen = (
            max_decoder_len if maxlen is None else min(maxlen, max_decoder_len)
        )
        self.stop_early = stop_early
        self.normalize_scores = normalize_scores
        self.len_penalty = len_penalty
        self.retain_dropout = retain_dropout
        self.sampling = sampling
        self.sampling_topk = sampling_topk
        self.sampling_temperature = sampling_temperature

    def generate_batch_translations(
        self,
        batch,
        maxlen_a=0.0,
        maxlen_b=None,
        prefix_size=0,
    ):
        """Yield individual translations of a batch.

        Args:
            batch (dict): The model input batch. Must have keys `net_input`, `target` and `ntokens`
            maxlen_a (float):
            maxlen_b (Optional[int]): Generate sequences of max lengths `maxlen_a*x + maxlen_b` where `x = input sentence length`
            prefix_size (int): Prefix size
        """
        if maxlen_b is None:
            maxlen_b = self.maxlen

        if "net_input" not in batch:
            return
        input = batch["net_input"]
        srclen = input["src_tokens"].size(1)
        with torch.no_grad():
            hypos = self.generate(
                input["src_tokens"],
                input["src_lengths"],
                maxlen=int(maxlen_a * srclen + maxlen_b),
                prefix_tokens=batch["target"][:, :prefix_size]
                if prefix_size > 0
                else None,
            )
        for i, id in enumerate(batch["id"].data):
            # remove padding
            src = strip_pad(input["src_tokens"].data[i, :], self.pad)
            ref = (
                strip_pad(batch["target"].data[i, :], self.pad)
                if batch["target"] is not None
                else None
            )
            yield id, src, ref, hypos[i]

    def generate(self, src_tokens, src_lengths, maxlen=None, prefix_tokens=None):
        """Generate a batch of translations."""
        with torch.no_grad():
            return self._generate(src_tokens, src_lengths, maxlen, prefix_tokens)

    def _generate(self, src_tokens, src_lengths, maxlen=None, prefix_tokens=None):
        bsz, srclen = src_tokens.size()
        maxlen = min(maxlen, self.maxlen) if maxlen is not None else self.maxlen

        # the max beam size is the dictionary size - 1, since we never select pad
        beam_size = self.beam_size
        beam_size = min(beam_size, self.vocab_size - 1)

        incremental_state = None
        if not self.retain_dropout:
            self.model.eval()
        if isinstance(self.model.decoder, TransformerDecoder):
            incremental_state = {}

        # compute the encoder output for each beam
        encoder_out = self.model.encoder(
            src_tokens.repeat(1, beam_size).view(-1, srclen),
            src_lengths.expand(beam_size, src_lengths.numel())
            .t()
            .contiguous()
            .view(-1),
        )

        # initialize buffers
        scores = src_tokens.data.new(bsz * beam_size, maxlen + 1).float().fill_(0)
        scores_buf = scores.clone()
        tokens = src_tokens.data.new(bsz * beam_size, maxlen + 2).fill_(self.pad)
        tokens_buf = tokens.clone()
        tokens[:, 0] = self.eos
        attn, attn_buf = None, None
        nonpad_idxs = None

        # list of completed sentences
        finalized = [[] for i in range(bsz)]
        finished = [False for i in range(bsz)]
        worst_finalized = [{"idx": None, "score": -math.inf} for i in range(bsz)]
        num_remaining_sent = bsz

        # number of candidate hypos per step
        cand_size = 2 * beam_size  # 2 x beam size in case half are EOS

        # offset arrays for converting between different indexing schemes
        bbsz_offsets = (torch.arange(0, bsz) * beam_size).unsqueeze(1).type_as(tokens)
        cand_offsets = torch.arange(0, cand_size).type_as(tokens)

        # helper function for allocating buffers on the fly
        buffers = {}

        def buffer(name, type_of=tokens):  # noqa
            if name not in buffers:
                buffers[name] = type_of.new()
            return buffers[name]

        def is_finished(sent, step, unfinalized_scores=None):
            """
            Check whether we've finished generation for a given sentence, by
            comparing the worst score among finalized hypotheses to the best
            possible score among unfinalized hypotheses.
            """
            assert len(finalized[sent]) <= beam_size
            if len(finalized[sent]) == beam_size:
                if self.stop_early or step == maxlen or unfinalized_scores is None:
                    return True
                # stop if the best unfinalized score is worse than the worst
                # finalized one
                best_unfinalized_score = unfinalized_scores[sent].max()
                if self.normalize_scores:
                    # We don't know why the reference adds 5 and divides by 6, perhaps for rounding
                    best_unfinalized_score /= ((maxlen + 5) / 6) ** self.len_penalty
                if worst_finalized[sent]["score"] >= best_unfinalized_score:
                    return True
            return False

        def finalize_hypos(step, bbsz_idx, eos_scores, unfinalized_scores=None):
            """
            Finalize the given hypotheses at this step, while keeping the total
            number of finalized hypotheses per sentence <= beam_size.
            Note: the input must be in the desired finalization order, so that
            hypotheses that appear earlier in the input are preferred to those
            that appear later.
            Args:
                step: current time step
                bbsz_idx: A vector of indices in the range [0, bsz*beam_size),
                    indicating which hypotheses to finalize
                eos_scores: A vector of the same size as bbsz_idx containing
                    scores for each hypothesis
                unfinalized_scores: A vector containing scores for all
                    unfinalized hypotheses
            """
            assert bbsz_idx.numel() == eos_scores.numel()

            # clone relevant token and attention tensors
            tokens_clone = tokens.index_select(0, bbsz_idx)
            tokens_clone = tokens_clone[
                :, 1 : step + 2
            ]  # skip the first index, which is EOS
            tokens_clone[:, step] = self.eos
            attn_clone = (
                attn.index_select(0, bbsz_idx)[:, :, 1 : step + 2]
                if attn is not None
                else None
            )

            # compute scores per token position
            pos_scores = scores.index_select(0, bbsz_idx)[:, : step + 1]
            pos_scores[:, step] = eos_scores
            # convert from cumulative to per-position scores
            pos_scores[:, 1:] = pos_scores[:, 1:] - pos_scores[:, :-1]

            # normalize sentence-level scores
            if self.normalize_scores:
                # We don't know why the reference adds 5 and divides by 6, perhaps for rounding
                eos_scores /= (((step + 1) + 5) / 6) ** self.len_penalty

            cum_unfin = []
            prev = 0
            for f in finished:
                if f:
                    prev += 1
                else:
                    cum_unfin.append(prev)

            sents_seen = set()
            for i, (idx, score) in enumerate(
                zip(bbsz_idx.tolist(), eos_scores.tolist())
            ):
                unfin_idx = idx // beam_size
                sent = unfin_idx + cum_unfin[unfin_idx]

                sents_seen.add((sent, unfin_idx))

                def get_hypo():

                    if attn_clone is not None:
                        # remove padding tokens from attn scores
                        hypo_attn = attn_clone[i][nonpad_idxs[sent]]
                        _, alignment = hypo_attn.max(dim=0)
                    else:
                        hypo_attn = None
                        alignment = None

                    return {
                        "tokens": tokens_clone[i],
                        "score": score,
                        "attention": hypo_attn,  # src_len x tgt_len
                        "alignment": alignment,
                        "positional_scores": pos_scores[i],
                    }

                if len(finalized[sent]) < beam_size:
                    finalized[sent].append(get_hypo())
                elif not self.stop_early and score > worst_finalized[sent]["score"]:
                    # replace worst hypo for this sentence with new/better one
                    worst_idx = worst_finalized[sent]["idx"]
                    if worst_idx is not None:
                        finalized[sent][worst_idx] = get_hypo()

                    # find new worst finalized hypo for this sentence
                    idx, s = min(
                        enumerate(finalized[sent]), key=lambda r: r[1]["score"]
                    )
                    worst_finalized[sent] = {
                        "score": s["score"],
                        "idx": idx,
                    }

            newly_finished = []
            for sent, unfin_idx in sents_seen:
                # check termination conditions for this sentence
                if not finished[sent] and is_finished(sent, step, unfinalized_scores):
                    finished[sent] = True
                    newly_finished.append(unfin_idx)
            return newly_finished

        reorder_state = None
        batch_idxs = None
        for step in range(maxlen + 1):  # one extra step for EOS marker
            # reorder decoder internal states based on the prev choice of beams
            # if 'MultiheadAttention.1.attn_state' in incremental_state:
            #     print(incremental_state['MultiheadAttention.1.attn_state'].keys())
            if reorder_state is not None:
                if batch_idxs is not None:
                    # update beam indices to take into account removed sentences
                    corr = batch_idxs - torch.arange(batch_idxs.numel()).type_as(
                        batch_idxs
                    )
                    reorder_state.view(-1, beam_size).add_(
                        corr.unsqueeze(-1) * beam_size
                    )
                if isinstance(self.model.decoder, TransformerDecoder):
                    self.model.decoder.reorder_incremental_state(
                        incremental_state, reorder_state
                    )
                encoder_out = self.model.encoder.reorder_encoder_out(
                    encoder_out, reorder_state
                )

            probs, avg_attn_scores = self._decode_one(
                tokens[:, : step + 1],
                self.model,
                encoder_out,
                incremental_state,
                log_probs=True,
            )
            if step == 0:
                # at the first step all hypotheses are equally likely, so use
                # only the first beam
                probs = probs.unfold(0, 1, beam_size).squeeze(2).contiguous()
                scores = scores.type_as(probs)
                scores_buf = scores_buf.type_as(probs)
            elif not self.sampling:
                # make probs contain cumulative scores for each hypothesis
                probs.add_(scores[:, step - 1].view(-1, 1))

            probs[:, self.pad] = -math.inf  # never select pad

            # Record attention scores
            if avg_attn_scores is not None:
                if attn is None:
                    attn = scores.new(bsz * beam_size, src_tokens.size(1), maxlen + 2)
                    attn_buf = attn.clone()
                    nonpad_idxs = src_tokens.ne(self.pad)
                attn[:, :, step + 1].copy_(avg_attn_scores)

            cand_scores = buffer("cand_scores", type_of=scores)
            cand_indices = buffer("cand_indices")
            cand_beams = buffer("cand_beams")
            eos_bbsz_idx = buffer("eos_bbsz_idx")
            eos_scores = buffer("eos_scores", type_of=scores)
            if step < maxlen:
                if prefix_tokens is not None and step < prefix_tokens.size(1):
                    probs_slice = probs.view(bsz, -1, probs.size(-1))[:, 0, :]
                    cand_scores = torch.gather(
                        probs_slice,
                        dim=1,
                        index=prefix_tokens[:, step].view(-1, 1).data,
                    ).expand(-1, cand_size)
                    cand_indices = (
                        prefix_tokens[:, step].view(-1, 1).expand(bsz, cand_size).data
                    )
                    cand_beams.resize_as_(cand_indices).fill_(0)
                elif self.sampling:
                    assert (
                        self.pad == 1
                    ), "sampling assumes the first two symbols can be ignored"

                    if self.sampling_topk > 0:
                        values, indices = probs[:, 2:].topk(self.sampling_topk)
                        exp_probs = values.div_(self.sampling_temperature).exp()
                        if step == 0:
                            torch.multinomial(
                                exp_probs, beam_size, replacement=True, out=cand_indices
                            )
                        else:
                            torch.multinomial(
                                exp_probs, 1, replacement=True, out=cand_indices
                            )
                        torch.gather(
                            exp_probs, dim=1, index=cand_indices, out=cand_scores
                        )
                        torch.gather(
                            indices, dim=1, index=cand_indices, out=cand_indices
                        )
                        cand_indices.add_(2)
                    else:
                        exp_probs = (
                            probs.div_(self.sampling_temperature)
                            .exp_()
                            .view(-1, self.vocab_size)
                        )

                        if step == 0:
                            # we exclude the first two vocab items, one of which is pad
                            torch.multinomial(
                                exp_probs[:, 2:],
                                beam_size,
                                replacement=True,
                                out=cand_indices,
                            )
                        else:
                            torch.multinomial(
                                exp_probs[:, 2:], 1, replacement=True, out=cand_indices
                            )

                        cand_indices.add_(2)
                        torch.gather(
                            exp_probs, dim=1, index=cand_indices, out=cand_scores
                        )

                    cand_scores.log_()
                    cand_indices = cand_indices.view(bsz, -1).repeat(1, 2)
                    cand_scores = cand_scores.view(bsz, -1).repeat(1, 2)
                    if step == 0:
                        cand_beams = torch.zeros(bsz, cand_size).type_as(cand_indices)
                    else:
                        cand_beams = (
                            torch.arange(0, beam_size)
                            .repeat(bsz, 2)
                            .type_as(cand_indices)
                        )
                        # make scores cumulative
                        cand_scores.add_(
                            torch.gather(
                                scores[:, step - 1].view(bsz, beam_size),
                                dim=1,
                                index=cand_beams,
                            )
                        )
                else:
                    # take the best 2 x beam_size predictions. We'll choose the first
                    # beam_size of these which don't predict eos to continue with.
                    torch.topk(
                        probs.view(bsz, -1),
                        k=min(
                            cand_size, probs.view(bsz, -1).size(1) - 1
                        ),  # -1 so we never select pad
                        out=(cand_scores, cand_indices),
                    )
                    torch.div(cand_indices, self.vocab_size, out=cand_beams)
                    cand_indices.fmod_(self.vocab_size)
            else:
                # finalize all active hypotheses once we hit maxlen
                # pick the hypothesis with the highest prob of EOS right now
                torch.sort(
                    probs[:, self.eos],
                    descending=True,
                    out=(eos_scores, eos_bbsz_idx),
                )
                num_remaining_sent -= len(
                    finalize_hypos(step, eos_bbsz_idx, eos_scores)
                )
                assert num_remaining_sent == 0
                break

            # cand_bbsz_idx contains beam indices for the top candidate
            # hypotheses, with a range of values: [0, bsz*beam_size),
            # and dimensions: [bsz, cand_size]
            cand_bbsz_idx = cand_beams.add(bbsz_offsets)

            # finalize hypotheses that end in eos
            eos_mask = cand_indices.eq(self.eos)

            finalized_sents = set()
            if step >= self.minlen:
                # only consider eos when it's among the top beam_size indices
                torch.masked_select(
                    cand_bbsz_idx[:, :beam_size],
                    mask=eos_mask[:, :beam_size],
                    out=eos_bbsz_idx,
                )
                if eos_bbsz_idx.numel() > 0:
                    torch.masked_select(
                        cand_scores[:, :beam_size],
                        mask=eos_mask[:, :beam_size],
                        out=eos_scores,
                    )
                    finalized_sents = finalize_hypos(
                        step, eos_bbsz_idx, eos_scores, cand_scores
                    )
                    num_remaining_sent -= len(finalized_sents)

            assert num_remaining_sent >= 0
            if num_remaining_sent == 0:
                break
            assert step < maxlen

            if len(finalized_sents) > 0:
                new_bsz = bsz - len(finalized_sents)

                # construct batch_idxs which holds indices of batches to keep for the next pass
                batch_mask = torch.ones(bsz).type_as(cand_indices)
                batch_mask[cand_indices.new(finalized_sents)] = 0
                batch_idxs = batch_mask.nonzero().squeeze(-1)

                eos_mask = eos_mask[batch_idxs]
                cand_beams = cand_beams[batch_idxs]
                bbsz_offsets.resize_(new_bsz, 1)
                cand_bbsz_idx = cand_beams.add(bbsz_offsets)

                cand_scores = cand_scores[batch_idxs]
                cand_indices = cand_indices[batch_idxs]
                if prefix_tokens is not None:
                    prefix_tokens = prefix_tokens[batch_idxs]

                scores = scores.view(bsz, -1)[batch_idxs].view(new_bsz * beam_size, -1)
                scores_buf.resize_as_(scores)
                tokens = tokens.view(bsz, -1)[batch_idxs].view(new_bsz * beam_size, -1)
                tokens_buf.resize_as_(tokens)
                if attn is not None:
                    attn = attn.view(bsz, -1)[batch_idxs].view(
                        new_bsz * beam_size, attn.size(1), -1
                    )
                    attn_buf.resize_as_(attn)
                bsz = new_bsz
            else:
                batch_idxs = None

            # set active_mask so that values > cand_size indicate eos hypos
            # and values < cand_size indicate candidate active hypos.
            # After, the min values per row are the top candidate active hypos
            active_mask = buffer("active_mask")
            torch.add(
                eos_mask.type_as(cand_offsets) * cand_size,
                cand_offsets[: eos_mask.size(1)],
                out=active_mask,
            )

            # get the top beam_size active hypotheses, which are just the hypos
            # with the smallest values in active_mask
            active_hypos, _ignore = buffer("active_hypos"), buffer("_ignore")
            torch.topk(
                active_mask,
                k=beam_size,
                dim=1,
                largest=False,
                out=(_ignore, active_hypos),
            )
            active_bbsz_idx = buffer("active_bbsz_idx")
            torch.gather(
                cand_bbsz_idx,
                dim=1,
                index=active_hypos,
                out=active_bbsz_idx,
            )
            active_scores = torch.gather(
                cand_scores,
                dim=1,
                index=active_hypos,
                out=scores[:, step].view(bsz, beam_size),
            )

            active_bbsz_idx = active_bbsz_idx.view(-1)
            active_scores = active_scores.view(-1)

            # copy tokens and scores for active hypotheses
            torch.index_select(
                tokens[:, : step + 1],
                dim=0,
                index=active_bbsz_idx,
                out=tokens_buf[:, : step + 1],
            )
            torch.gather(
                cand_indices,
                dim=1,
                index=active_hypos,
                out=tokens_buf.view(bsz, beam_size, -1)[:, :, step + 1],
            )
            if step > 0:
                torch.index_select(
                    scores[:, :step],
                    dim=0,
                    index=active_bbsz_idx,
                    out=scores_buf[:, :step],
                )
            torch.gather(
                cand_scores,
                dim=1,
                index=active_hypos,
                out=scores_buf.view(bsz, beam_size, -1)[:, :, step],
            )

            # copy attention for active hypotheses
            if attn is not None:
                torch.index_select(
                    attn[:, :, : step + 2],
                    dim=0,
                    index=active_bbsz_idx,
                    out=attn_buf[:, :, : step + 2],
                )

            # swap buffers
            tokens, tokens_buf = tokens_buf, tokens
            scores, scores_buf = scores_buf, scores
            if attn is not None:
                attn, attn_buf = attn_buf, attn

            # reorder incremental state in decoder
            reorder_state = active_bbsz_idx

        # sort by score descending
        for sent in range(len(finalized)):
            finalized[sent] = sorted(
                finalized[sent], key=lambda r: r["score"], reverse=True
            )

        return finalized

    def _decode_one(self, tokens, model, encoder_out, incremental_state, log_probs):
        with torch.no_grad():
            if incremental_state is not None:
                decoder_out = list(
                    model.decoder(
                        tokens, encoder_out, incremental_state=incremental_state
                    )
                )
            else:
                decoder_out = list(model.decoder(tokens, encoder_out))
            decoder_out[0] = decoder_out[0][:, -1, :]
            attn = decoder_out[1]
            if attn is not None:
                attn = attn[:, -1, :]
        probs = get_normalized_probs(decoder_out, log_probs=log_probs)
        return probs, attn

    def translate_batch(
        self,
        batch,
        maxlen_a=1.0,
        maxlen_b=50,
        prefix_size=0,
        remove_bpe=None,
        nbest=1,
        ignore_case=True,
    ):
        """
        Args:
            batch (dict): The model input batch. Must have keys `net_input`, `target` and `ntokens`
            maxlen_a (float): Default 1.0
            maxlen_b (Optional[int]): Generate sequences of max lengths `maxlen_a*x + maxlen_b` where `x = input sentence length`.
                Default 50
            prefix_size (int): Prefix size. Default 0
            remove_bpe (Optional[str]): BPE token. Default `None`
            nbest (int): Number of hypotheses to output. Default 1
            ignore_case (bool): Ignore case druing online eval. Default `True`

        Returns:
            (list[str], list[str]): The translations and their targets for the given batch
        """
        translations = self.generate_batch_translations(
            batch,
            maxlen_a=maxlen_a,
            maxlen_b=maxlen_b,
            prefix_size=prefix_size,
        )

        ref_toks = []
        sys_toks = []
        for sample_id, src_tokens, target_tokens, hypos in translations:
            # Process input and ground truth
            has_target = target_tokens is not None
            target_tokens = target_tokens.int().cpu() if has_target else None

            src_str = self.src_dict.string(src_tokens, remove_bpe)
            if has_target:
                target_str = self.trg_dict.string(target_tokens, remove_bpe)

            # Process top predictions
            for i, hypo in enumerate(hypos[: min(len(hypos), nbest)]):
                hypo_tokens, hypo_str, alignment = post_process_prediction(
                    hypo_tokens=hypo["tokens"].int().cpu(),
                    alignment=hypo["alignment"].int().cpu()
                    if hypo["alignment"] is not None
                    else None,
                    align_dict=None,
                    tgt_dict=self.trg_dict,
                    remove_bpe=remove_bpe,
                )

                # Score only the top hypothesis
                if has_target and i == 0:
                    src_str = detokenize_subtokenized_sentence(src_str)
                    target_str = detokenize_subtokenized_sentence(target_str)
                    hypo_str = detokenize_subtokenized_sentence(hypo_str)
                    sys_tok = bleu_tokenize(
                        hypo_str.lower() if ignore_case else hypo_str
                    )
                    ref_tok = bleu_tokenize(
                        target_str.lower() if ignore_case else target_str
                    )
                    sys_toks.append(sys_tok)
                    ref_toks.append(ref_tok)

        return sys_toks, ref_toks


def get_normalized_probs(net_output, log_probs):
    """Get normalized probabilities (or log probs) from a net's output."""
    logits = net_output[0]
    if log_probs:
        return F.log_softmax(logits, dim=-1, dtype=torch.float32)
    else:
        return F.softmax(logits, dim=-1, dtype=torch.float32)



import numpy as np
import torch
import torch.distributed as dist

from mlbench_core.utils.pytorch.utils import orthogonalize, pack_tensors, unpack_tensors

try:
    import horovod.torch as hvd
except ImportError as e:
    pass

from mlbench_core.aggregation.pytorch.aggregation import Aggregation

AVG_WORLD = "avg_world"
AVG_CUSTOM = "avg_custom"
ALLREDUCE_AGGREGATION_OPS = [AVG_WORLD, AVG_CUSTOM]
"""
All possible aggregations for AllReduceAggregation
"""


class AllReduceAggregation(Aggregation):
    """Aggregate weights / models from different processes using all-reduce aggregation

    Args:
        world_size (int): Current distributed world size
        divide_before (bool): Perform division before reduction (avoid overflow)
        use_cuda (bool): Use cuda tensors for reduction
    """

    def __init__(self, world_size, divide_before=False, use_cuda=False):
        self.world_size = world_size
        self.divide_before = divide_before
        super(AllReduceAggregation, self).__init__(use_cuda=use_cuda)

    def _reduce(self, data):
        """Reduces the given tensor using `torch.distributed` and op=`dist.ReduceOp.SUM`

        Args:
            data (:obj:`torch.Tensor`): The tensor to reduce

        Returns:
            (:obj:`torch.Tensor`): The reduced Tensor
        """
        dist.all_reduce(data, op=dist.ReduceOp.SUM)
        return data

    def _divide(self, data, op, denom=None):
        """Divides the given `data` tensor by

        - `world_size` if op == `avg_world`
        - `denom`, if op == `custom_avg`

        Args:
            data (:obj:`torch.Tensor`): Data tensor to divide
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)

        Returns:
            (:obj:`torch.Tensor`): The resulting tensor
        """
        if op not in ALLREDUCE_AGGREGATION_OPS:
            raise NotImplementedError("Allreduce not implemented for op={}".format(op))

        if op == AVG_WORLD:
            data.div_(self.world_size)
        elif op == AVG_CUSTOM:
            if denom is None or denom == 0:
                raise ValueError("Denominator should be one element tensor")
            data.div_(denom)

        return data

    def _agg(self, data, op, denom=None):
        """Aggregate data using `op` operation.

        - If op == `avg_world`, the reduced tensor will be divided by `world_size`,
        - if op == `custom_avg`, the reduced tensor will be divided by `denom`

        If `self.divide_before`, the division is performed before reduction.
        This can be helpful to avoid overflows when using `float16` training

        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggregated.
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)

        Returns:
            (:obj:`torch.Tensor`): The aggregated tensor.
        """
        if self.divide_before:
            data = self._divide(data, op, denom)

        if self.world_size > 1:
            data = self._reduce(data)
            if dist.get_backend() == dist.Backend.NCCL:
                torch.cuda.synchronize()

        if not self.divide_before:
            data = self._divide(data, op, denom)
        return data


class AllReduceAggregationHVD(AllReduceAggregation):
    """Implements `AllReduceAggregation` using horovod for communication"""

    def _reduce(self, data):
        data = hvd.allreduce(data, op=hvd.Sum)
        return data


class SparsifiedAggregation(Aggregation):
    """Aggregate sparsified updates."""

    def __init__(self, model, use_cuda=False):
        super(SparsifiedAggregation, self).__init__(use_cuda=use_cuda)
        pass

    def _agg(self, data, op, denom=None):
        pass


class PowerAggregation(Aggregation):
    """Aggregate updates using power iteration and error feedback.

    Args:
            model (:obj:`nn.Module`): Model which contains parameters for SGD
            use_cuda (bool): Whether to use cuda tensors for aggregation
            reuse_query (bool): Whether to use warm start to initialize the power iteration
            rank (int): The rank of the gradient approximation
    """

    def __init__(self, model, use_cuda=False, reuse_query=False, rank=1):
        super(PowerAggregation, self).__init__(use_cuda=use_cuda)
        self.p_memory = None
        self.q_memory = None
        self.reuse_query = reuse_query
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.rng = np.random.RandomState(1)
        self.n_workers = dist.get_world_size()
        self.rank = rank
        self.memories = [torch.zeros_like(param) for param in model.parameters()]
        self.send_buffers = [torch.zeros_like(param) for param in model.parameters()]

    def set_random(self, vector):
        """Sets the data in `vector` to random values."""
        torch.manual_seed(self.rng.randint(1_000_000_000))
        vector.data[:] = torch.randn(*vector.shape, device=self.device)

    def _agg_gradients_by_model(self, model, op, denom=None):
        """Aggregate models gradients, all layers at once

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.
            denom (None): Not used here

        """
        grads = [t.grad.data for t in model.parameters()]
        aggregated = self._agg(grads, op=op)

        for i, param in enumerate(model.parameters()):
            param.grad.data = aggregated[i]

    def agg_weights(self, by_layer=False):
        raise NotImplementedError("PowerSGD doesn't allow aggregation by weights")

    def agg_model(self, by_layer=False):
        if by_layer:
            raise NotImplementedError("PowerSGD doesn't allow aggregation by layer")
        else:
            return self._agg_gradients_by_model

    def _agg(self, data, op, denom=None):
        """Aggregate data using `op` operation.

        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggragated.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.
            denom (None): Not used here

        Returns:
            :obj:`torch.Tensor`: An aggregated tensor.
        """
        if op == "avg":
            for grad, memory, send_bfr in zip(data, self.memories, self.send_buffers):
                send_bfr.data[:] = grad + memory
            self.reduce(self.send_buffers, data, self.memories)
        else:
            raise NotImplementedError("op {} is not supported yet.".format(op))
        return data

    def reduce(self, grad_in, grad_out, memory_out):
        """Reduces the gradients between the workers in place and calculates error feedback.

        Args:
            grad_in (list[torch.Tensor]): The gradients to reduce.
            grad_out (list[torch.Tensor]): Used for storing the reduced gradients.
            memory_out (list[torch.Tensor]): Used for storing error feedback.
        """
        # Split the tensors into rank1-ones that will be reduced un-compressed
        # and rank > 1 tensors that are compressed
        rank1_tensors = [
            (tensor, out, mem)
            for tensor, out, mem in zip(grad_in, grad_out, memory_out)
            if tensor.ndimension() <= 1
        ]
        high_rank_tensors = [
            (tensor, out, mem)
            for tensor, out, mem in zip(grad_in, grad_out, memory_out)
            if tensor.ndimension() > 1
        ]

        # We are building a rank-1 approximation of every tensor
        # that can be interpreted as a matrix. Let the approximation be
        # M = p q^T
        # We are allocating consequtive memory for the p's and q's

        memory_is_uninitialized = self.p_memory is None

        p_total_size = 0
        q_total_size = 0
        for tensor, _, _ in high_rank_tensors:
            matrix = tensor.view(tensor.shape[0], -1)
            n, m = matrix.shape
            rank = min(n, m, self.rank)
            p_total_size += n * rank
            q_total_size += m * rank
        if self.p_memory is None:
            self.p_memory = torch.empty(p_total_size, device=self.device)
            self.q_memory = torch.empty(q_total_size, device=self.device)

        # Find them again and make lists of pointers
        ps = []
        qs = []
        p_idx = 0
        q_idx = 0
        for tensor, _, _ in high_rank_tensors:
            matrix = tensor.view(tensor.shape[0], -1)
            n, m = matrix.shape
            rank = min(n, m, self.rank)
            ps.append(self.p_memory[p_idx : p_idx + n * rank].view(n, rank))
            qs.append(self.q_memory[q_idx : q_idx + m * rank].view(m, rank))
            p_idx += n * rank
            q_idx += m * rank

        for (tensor, _, _), q, p in zip(high_rank_tensors, qs, ps):
            matrix = tensor.view(tensor.shape[0], -1)
            n, m = matrix.shape

            if self.reuse_query and not memory_is_uninitialized:
                pass
            else:
                # Sample a query vector q
                self.set_random(q)

        for (tensor, _, _), q, p in zip(high_rank_tensors, qs, ps):
            matrix = tensor.view(tensor.shape[0], -1)
            torch.matmul(matrix, q, out=p)

        dist.all_reduce(self.p_memory)

        # Start communicating rank 1 tensors
        rank1_packed, rank1_indices, rank1_sizes = pack_tensors(
            [tensor for (tensor, _, _) in rank1_tensors]
        )

        rank1_handle = dist.all_reduce(rank1_packed, async_op=True)

        for p in ps:
            orthogonalize(p)

        for p, q, (tensor, _, _) in zip(ps, qs, high_rank_tensors):
            matrix = tensor.view(tensor.shape[0], -1)
            torch.matmul(matrix.t(), p, out=q)

        dist.all_reduce(self.q_memory)
        self.q_memory.data[:] /= self.n_workers

        for p, q, (tensor, out, mem) in zip(ps, qs, high_rank_tensors):
            # Set the output gradient
            torch.matmul(p, q.t(), out=out.data[:])
            mem.data[:] = tensor - out

        rank1_handle.wait()
        rank1_packed /= self.n_workers
        rank1_unpacked = unpack_tensors(rank1_packed, rank1_indices, rank1_sizes)
        for i, (_, out, _) in enumerate(rank1_tensors):
            out[:] = rank1_unpacked[i]
